from ocp_resources.resource import NamespacedResource


class ImageStreamTag(NamespacedResource):
    """
    ImageStreamTag object.
    """

    api_group = NamespacedResource.ApiGroup.IMAGE_OPENSHIFT_IO



# -*- coding: utf-8 -*-

from ocp_resources.cluster_role import ClusterRole
from ocp_resources.resource import Resource


class ClusterRoleBinding(Resource):
    """
    ClusterRoleBinding object.
    """

    api_group = Resource.ApiGroup.RBAC_AUTHORIZATION_K8S_IO

    def __init__(
        self,
        name=None,
        cluster_role=None,
        subjects=None,
        **kwargs,
    ):
        super().__init__(name=name, **kwargs)
        self.cluster_role = cluster_role
        self.subjects = subjects

    def to_dict(self):
        self.res = super().to_dict()

        self.res.setdefault("roleRef", {})
        self.res["roleRef"] = {
            "apiGroup": self.api_group,
            "kind": ClusterRole.kind,
            "name": self.cluster_role,
        }

        if self.subjects:
            self.res.setdefault("subjects", self.subjects)

        return self.res



import time

import yaml

from ocp_resources.logger import get_logger


LOGGER = get_logger(name=__name__)


class TimeoutExpiredError(Exception):
    def __init__(self, value):
        super().__init__()
        self.value = value

    def __str__(self):
        return f"Timed Out: {self.value}"


class TimeoutSampler:
    """
    Samples the function output.

    This is a generator object that at first yields the output of function
    `func`. After the yield, it either raises instance of `TimeoutExpiredError` or
    sleeps `sleep` seconds.

    Yielding the output allows you to handle every value as you wish.

    exceptions_dict should be in the following format:
    {
        exception0: [exception0_msg0],
        exception1: [
            exception1_msg0,
            exception1_msg1
        ],
        exception2: []
    }

    If an exception is raised within `func`:
        Example exception inheritance:
            class Exception
            class AExampleError(Exception)
            class BExampleError(AExampleError)

        The raised exception's class will fall into one of three categories:
            1. An exception class specifically declared in exceptions_dict
                exceptions_dict: {BExampleError: []}
                raise: BExampleError
                result: continue

            2. A child class inherited from an exception class in exceptions_dict
                exceptions_dict: {AExampleError: []}
                raise: BExampleError
                result: continue

            3. Everything else, this will always re-raise the exception
                exceptions_dict: {BExampleError: []}
                raise: AExampleError
                result: raise

    Args:
        wait_timeout (int): Time in seconds to wait for func to return a value equating to True
        sleep (int): Time in seconds between calls to func
        func (function): to be wrapped by TimeoutSampler
        exceptions_dict (dict): Exception handling definition
        print_log (bool): Print elapsed time to log
    """

    def __init__(
        self,
        wait_timeout,
        sleep,
        func,
        exceptions_dict=None,
        print_log=True,
        **func_kwargs,
    ):
        self.wait_timeout = wait_timeout
        self.sleep = sleep
        self.func = func
        self.func_kwargs = func_kwargs
        self.elapsed_time = None
        self.print_log = print_log

        self.exceptions_dict = exceptions_dict or {Exception: []}
        self._exceptions = tuple(self.exceptions_dict.keys())

    def _get_func_info(self, _func, type_):
        # If func is partial function.
        if getattr(_func, "func", None):
            return self._get_func_info(_func=_func.func, type_=type_)

        res = getattr(_func, type_, None)
        if res:
            # If func is lambda function.
            if _func.__name__ == "<lambda>":
                if type_ == "__module__":
                    return f"{res}.{_func.__qualname__.split('.')[1]}"

                elif type_ == "__name__":
                    free_vars = _func.__code__.co_freevars
                    free_vars = f"{'.'.join(free_vars)}." if free_vars else ""
                    return f"lambda: {free_vars}{'.'.join(_func.__code__.co_names)}"
            return res

    @property
    def _func_log(self):
        """
        Returns:
            string: `func` information to include in log message
        """
        _func_kwargs = f"Kwargs: {self.func_kwargs}" if self.func_kwargs else ""
        _func_module = self._get_func_info(_func=self.func, type_="__module__")
        _func_name = self._get_func_info(_func=self.func, type_="__name__")
        return f"Function: {_func_module}.{_func_name} {_func_kwargs}".strip()

    def __iter__(self):
        """
        Iterator

        Yields:
            any: Return value from `func`
        """
        timeout_watch = TimeoutWatch(timeout=self.wait_timeout)
        if self.print_log:
            LOGGER.info(
                f"Waiting for {self.wait_timeout} seconds, retry every {self.sleep} seconds. ({self._func_log})"
            )

        last_exp = None
        while timeout_watch.remaining_time() > 0:
            try:
                self.elapsed_time = self.wait_timeout - timeout_watch.remaining_time()
                yield self.func(**self.func_kwargs)
                self.elapsed_time = None

                time.sleep(self.sleep)

            except self._exceptions as exp:
                last_exp = exp
                last_exp_log = self._get_exception_log(exp=last_exp)
                if self._is_raisable_exception(exp=last_exp):
                    LOGGER.error(last_exp_log)
                    raise exp

                self.elapsed_time = None
                time.sleep(self.sleep)

            finally:
                if self.elapsed_time and self.print_log:
                    LOGGER.info(f"Elapsed time: {self.elapsed_time}")

        raise TimeoutExpiredError(self._get_exception_log(exp=last_exp))

    @staticmethod
    def _is_exception_matched(exp, exception_messages):
        """
        Args:
            exp (Exception): Exception object raised by `func`
            exception_messages (list): Either an empty list allowing all text,
                                       or a list of allowed strings to match against the exception text.

        Returns:
            bool: True if exception text is allowed or no exception text given, False otherwise
        """
        if not exception_messages:
            return True

        # Prevent match if provided with empty string
        return any(msg and msg in str(exp) for msg in exception_messages)

    def _is_raisable_exception(self, exp):
        """
        Verify whether exception should be raised during execution of `func`

        Args:
            exp (Exception): Exception object raised by `func`

        Returns:
            bool: True if exp should be raised, False otherwise
        """

        for entry in self.exceptions_dict:
            if isinstance(exp, entry):  # Check inheritance for raised exception
                exception_messages = self.exceptions_dict.get(entry)
                if self._is_exception_matched(
                    exp=exp, exception_messages=exception_messages
                ):
                    return False

        return True

    def _get_exception_log(self, exp):
        """
        Args:
            exp (any): Raised exception

        Returns:
            string: Log message for exception
        """
        exp_name = exp.__class__.__name__ if exp else "N/A"

        last_exception_log = f"Last exception: {exp_name}: {exp}"
        return f"{self.wait_timeout}\n{self._func_log}\n{last_exception_log}"


class TimeoutWatch:
    """
    A time counter allowing to determine the time remaining since the start
    of a given interval
    """

    def __init__(self, timeout):
        self.timeout = timeout
        self.start_time = time.time()

    def remaining_time(self):
        """
        Return the remaining part of timeout since the object was created.
        """
        return self.start_time + self.timeout - time.time()


def skip_existing_resource_creation_teardown(
    resource, export_str, user_exported_args, check_exists=True
):
    """
    Args:
        resource (Resource): Resource to match against.
        export_str (str): The user export str. (REUSE_IF_RESOURCE_EXISTS or SKIP_RESOURCE_TEARDOWN)
        user_exported_args (str): Value of export_str. (os.environ.get)
        check_exists (bool): Check if resource exists before return. (applied only for REUSE_IF_RESOURCE_EXISTS)

    Returns:
        Resource or None: If resource match.
    """

    def _return_resource(_resource, _check_exists, _msg):
        """
        Return the resource if exists when got _check_exists else return None.
        If _check_exists=False returns the resource.
        """
        if not _check_exists:  # In case of SKIP_RESOURCE_TEARDOWN
            return _resource

        elif _resource.exists:  # In case of REUSE_IF_RESOURCE_EXISTS
            LOGGER.warning(_msg)
            return _resource

    resource_to_dict = resource.to_dict()
    resource_name = resource_to_dict["metadata"]["name"]
    resource_namespace = resource_to_dict["metadata"].get("namespace")
    skip_create_warn_msg = (
        f"Skip resource {resource.kind} {resource_name} creation, using existing one."
        f" Got {export_str}={user_exported_args}"
    )
    user_args = yaml.safe_load(stream=user_exported_args)
    if resource.kind in user_args:
        _resource_args = user_args[resource.kind]
        if not _resource_args:
            # Match only by kind, user didn't send name and/or namespace.
            return _return_resource(
                _resource=resource,
                _check_exists=check_exists,
                _msg=skip_create_warn_msg,
            )

        for _name, _namespace in _resource_args.items():
            if resource_name == _name and (
                resource_namespace == _namespace
                or not (resource_namespace and _namespace)
            ):
                return _return_resource(
                    _resource=resource,
                    _check_exists=check_exists,
                    _msg=skip_create_warn_msg,
                )



# -*- coding: utf-8 -*-


from ocp_resources.constants import PROTOCOL_ERROR_EXCEPTION_DICT, TIMEOUT_4MINUTES
from ocp_resources.logger import get_logger
from ocp_resources.resource import NamespacedResource
from ocp_resources.utils import TimeoutSampler
from ocp_resources.virtual_machine_instance import VirtualMachineInstance


LOGGER = get_logger(name=__name__)


class VirtualMachine(NamespacedResource):
    """
    Virtual Machine object, inherited from Resource.
    Implements actions start / stop / status / wait for VM status / is running
    """

    api_group = NamespacedResource.ApiGroup.KUBEVIRT_IO

    class RunStrategy:
        MANUAL = "Manual"
        HALTED = "Halted"
        ALWAYS = "Always"
        RERUNONFAILURE = "RerunOnFailure"

    class Status(NamespacedResource.Status):
        MIGRATING = "Migrating"
        PAUSED = "Paused"
        PROVISIONING = "Provisioning"
        STARTING = "Starting"
        STOPPED = "stopped"
        STOPPING = "Stopping"

    def __init__(
        self,
        name=None,
        namespace=None,
        client=None,
        body=None,
        teardown=True,
        privileged_client=None,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        **kwargs,
    ):
        super().__init__(
            name=name,
            namespace=namespace,
            client=client,
            teardown=teardown,
            privileged_client=privileged_client,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )
        self.body = body

    @property
    def _subresource_api_url(self):
        return (
            f"{self.client.configuration.host}/"
            f"apis/subresources.kubevirt.io/{self.api.api_version}/"
            f"namespaces/{self.namespace}/virtualmachines/{self.name}"
        )

    def api_request(self, method, action, **params):
        return super().api_request(
            method=method, action=action, url=self._subresource_api_url, **params
        )

    def to_dict(self):
        res = super().to_dict()
        if self.yaml_file:
            return res

        body_spec = self.body.get("spec") if self.body else None
        res["spec"] = body_spec or {"template": {"spec": {}}}
        return res

    def start(self, timeout=TIMEOUT_4MINUTES, wait=False):
        self.api_request(method="PUT", action="start")
        if wait:
            return self.wait_for_status(timeout=timeout, status=True)

    def restart(self, timeout=TIMEOUT_4MINUTES, wait=False):
        self.api_request(method="PUT", action="restart")
        if wait:
            self.vmi.virt_launcher_pod.wait_deleted()
            return self.vmi.wait_until_running(timeout=timeout, stop_status="dummy")

    def stop(
        self, timeout=TIMEOUT_4MINUTES, vmi_delete_timeout=TIMEOUT_4MINUTES, wait=False
    ):
        self.api_request(method="PUT", action="stop")
        if wait:
            self.wait_for_status(timeout=timeout, status=None)
            return self.vmi.wait_deleted(timeout=vmi_delete_timeout)

    def wait_for_status(self, status, timeout=TIMEOUT_4MINUTES, sleep=1):
        """
        Wait for resource to be in status

        Args:
            status: Expected status: True for a running VM, None for a stopped VM.
            timeout (int): Time to wait for the resource.

        Raises:
            TimeoutExpiredError: If timeout reached.
        """
        LOGGER.info(
            f"Wait for {self.kind} {self.name} status to be {'ready' if status == True else status}"
        )
        samples = TimeoutSampler(
            wait_timeout=timeout,
            sleep=sleep,
            exceptions_dict=PROTOCOL_ERROR_EXCEPTION_DICT,
            func=self.api.get,
            field_selector=f"metadata.name=={self.name}",
            namespace=self.namespace,
        )
        for sample in samples:
            if sample.items and self.ready == status:
                # VM with runStrategy does not have spec.running attribute
                # VM status should be taken from spec.status.ready
                return

    def get_interfaces(self):
        return self.instance.spec.template.spec.domain.devices.interfaces

    @property
    def vmi(self):
        """
        Get VMI

        Returns:
            VirtualMachineInstance: VMI
        """
        return VirtualMachineInstance(
            client=self.client,
            name=self.name,
            namespace=self.namespace,
            privileged_client=self.privileged_client or self.client,
        )

    @property
    def ready(self):
        """
        Get VM status

        Returns:
            True if Running else None
        """
        return self.instance.get("status", {}).get("ready")

    @property
    def printable_status(self):
        """
        Get VM printableStatus

        Returns:
            VM printableStatus if VM.status.printableStatus else None
        """
        return self.instance.get("status", {}).get("printableStatus")

    def wait_for_status_none(self, status, timeout=TIMEOUT_4MINUTES):
        LOGGER.info(f"Wait for {self.kind} {self.name} status {status} to be None")
        for sample in TimeoutSampler(
            wait_timeout=timeout,
            sleep=1,
            exceptions_dict=PROTOCOL_ERROR_EXCEPTION_DICT,
            func=lambda: self.instance.get("status", {}).get(status),
        ):
            if sample is None:
                return



from ocp_resources.constants import TIMEOUT_4MINUTES
from ocp_resources.resource import NamespacedResource


class ImageStream(NamespacedResource):
    """
    ImageStream object.
    """

    api_group = NamespacedResource.ApiGroup.IMAGE_OPENSHIFT_IO

    def __init__(
        self,
        name=None,
        namespace=None,
        client=None,
        lookup_policy=False,
        tags=None,
        teardown=True,
        privileged_client=None,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        **kwargs,
    ):
        super().__init__(
            name=name,
            namespace=namespace,
            client=client,
            teardown=teardown,
            privileged_client=privileged_client,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )
        self.tags = tags
        self.lookup_policy = lookup_policy

    def to_dict(self):
        res = super().to_dict()
        if self.yaml_file:
            return res
        res.update(
            {"spec": {"lookupPolicy": {"local": self.lookup_policy}, "tags": self.tags}}
        )
        return res



from ocp_resources.constants import TIMEOUT_4MINUTES
from ocp_resources.resource import Resource


class PriorityClass(Resource):
    """
    Priority Class object.
    """

    api_group = Resource.ApiGroup.SCHEDULING_K8S_IO

    def __init__(
        self,
        name=None,
        client=None,
        teardown=True,
        value=None,
        global_default=False,
        description=None,
        preemption_policy=None,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        **kwargs,
    ):
        super().__init__(
            name=name,
            client=client,
            teardown=teardown,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )
        self.value = value
        self.global_default = global_default
        self.description = description
        self.preemption_policy = preemption_policy

    def to_dict(self):
        res = super().to_dict()
        if self.yaml_file:
            return res

        if self.value:
            res["value"] = self.value
        if self.global_default:
            res["globalDefault"] = self.global_default
        if self.description:
            res["description"] = self.description
        if self.preemption_policy:
            res["preemptionPolicy"] = self.preemption_policy
        return res



from ocp_resources.constants import TIMEOUT_4MINUTES
from ocp_resources.resource import NamespacedResource


class MachineHealthCheck(NamespacedResource):
    """
    MachineHealthCheck object.
    """

    api_group = NamespacedResource.ApiGroup.MACHINE_OPENSHIFT_IO

    def __init__(
        self,
        name=None,
        namespace=None,
        cluster_name=None,
        machineset_name=None,
        client=None,
        machine_role="worker",
        machine_type="worker",
        node_startup_timeout="120m",
        max_unhealthy=2,  # can also be a persentage, for e.g. "40%"
        unhealthy_timeout="300s",
        reboot_strategy=False,
        teardown=True,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        **kwargs,
    ):
        super().__init__(
            name=name,
            namespace=namespace,
            client=client,
            teardown=teardown,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )
        self.cluster_name = cluster_name
        self.machineset_name = machineset_name
        self.machine_role = machine_role
        self.machine_type = machine_type
        self.node_startup_timeout = node_startup_timeout
        self.max_unhealthy = max_unhealthy
        self.unhealthy_timeout = unhealthy_timeout
        self.reboot_strategy = reboot_strategy

    def to_dict(self):
        res = super().to_dict()
        if self.yaml_file:
            return res

        if self.reboot_strategy:
            res["metadata"]["annotations"] = {
                f"{self.api_group}/remediation-strategy": "external-baremetal"
            }
        res.setdefault("spec", {})
        res["spec"]["nodeStartupTimeout"] = self.node_startup_timeout
        res["spec"]["maxUnhealthy"] = self.max_unhealthy
        res["spec"].setdefault("selector", {})
        res["spec"]["selector"]["matchLabels"] = {
            f"{self.api_group}/cluster-api-cluster": self.cluster_name,
            f"{self.api_group}/cluster-api-machine-role": self.machine_role,
            f"{self.api_group}/cluster-api-machine-type": self.machine_type,
            f"{self.api_group}/cluster-api-machineset": self.machineset_name,
        }
        res["spec"]["unhealthyConditions"] = [
            {
                "type": self.Condition.READY,
                "timeout": self.unhealthy_timeout,
                "status": self.Condition.Status.FALSE,
            },
            {
                "type": self.Condition.READY,
                "timeout": self.unhealthy_timeout,
                "status": self.Condition.Status.UNKNOWN,
            },
        ]
        return res



# -*- coding: utf-8 -*-

from ocp_resources.resource import NamespacedResource


class Role(NamespacedResource):
    """
    Role object.
    """

    api_group = NamespacedResource.ApiGroup.RBAC_AUTHORIZATION_K8S_IO



import re

import jinja2
import yaml


class MissingTemplateVariables(Exception):
    def __init__(self, var, template):
        self.var = var
        self.template = template

    def __str__(self):
        return f"Missing variables {self.var} for template {self.template}"


def generate_yaml_from_template(**kwargs):
    """
    Generate JSON from yaml file_

    Keyword Args:
        name (str):
        image (str):

    Returns:
        dict: Generated from template file

    Raises:
        MissingTemplateVariables: If not all template variables exists

    Examples:
        generate_yaml_from_template(file_='path/to/file/name', name='vm-name-1')
    """
    file_ = "tests/manifests/vm.yaml"
    with open(file_, "r") as stream:
        data = stream.read()

    # Find all template variables
    template_vars = [i.split()[1] for i in re.findall(r"{{ .* }}", data)]
    for var in template_vars:
        if var not in kwargs.keys():
            raise MissingTemplateVariables(var=var, template=file_)

    template = jinja2.Template(source=data)
    out = template.render(**kwargs)
    return yaml.safe_load(stream=out)



from ocp_resources.constants import TIMEOUT_4MINUTES
from ocp_resources.resource import NamespacedResource


class KubeDescheduler(NamespacedResource):
    api_group = NamespacedResource.ApiGroup.OPERATOR_OPENSHIFT_IO

    def __init__(
        self,
        name=None,
        namespace=None,
        profiles=None,
        descheduling_interval=3600,
        log_level="Normal",
        managemet_state="Managed",
        operator_log_level="Normal",
        teardown=True,
        client=None,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        **kwargs,
    ):
        """
        Create Descheduler object.

        Args:
            profiles (list): list of descheduling profiles. Supported:
                - AffinityAndTaints
                - TopologyAndDuplicates
                - LifecycleAndUtilization
            log_level (str): logging of a component. Supported: "Normal", "Debug", "Trace", "TraceAll"
            operator_log_level (str): logging of an operator. Supported: "Normal", "Debug", "Trace", "TraceAll"
        """
        super().__init__(
            name=name,
            namespace=namespace,
            client=client,
            teardown=teardown,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )
        self.profiles = profiles or ["AffinityAndTaints"]
        self.descheduling_interval = descheduling_interval
        self.log_level = log_level
        self.managemet_state = managemet_state
        self.operator_log_level = operator_log_level

    def to_dict(self):
        res = super().to_dict()
        if self.yaml_file:
            return res

        res.update(
            {
                "spec": {
                    "deschedulingIntervalSeconds": self.descheduling_interval,
                    "logLevel": self.log_level,
                    "managementState": self.managemet_state,
                    "operatorLogLevel": self.operator_log_level,
                    "profiles": self.profiles,
                }
            }
        )
        return res



# This test uses https://github.com/RedHatQE/openshift-resources-definitions to get resource definitions json file
# File: resources_definitions.json

import ast
import json
import os
import re

import pytest
import requests


def _api_group_name(api_value):
    return re.sub("_", ".", api_value.lower())


def _api_group_dict(resource_dict, api_group_name):
    resource_dict_api_group = resource_dict["api_group"]
    api_group = resource_dict_api_group.get(api_group_name)
    if not api_group:
        try:
            api_group = resource_dict_api_group["null"]
        except KeyError:
            api_group = resource_dict_api_group[[*resource_dict_api_group][0]]

    return api_group


def _process_api_type(api_type, api_value, resource_dict, cls):
    if api_type == "api_group":
        return _get_api_group(
            api_value=api_value,
            cls=cls,
            resource_dict=resource_dict,
        )

    if api_type == "api_version":
        return _get_api_version(
            api_value=api_value,
            cls=cls,
            resource_dict=resource_dict,
        )

    return [f"Resource {cls.name} must have api_group or api_version"]


def _get_api_group_and_version(bodies):
    for targets in bodies:
        api_type = targets.targets[0].id
        return api_type, getattr(
            targets.value, "attr", getattr(targets.value, "s", None)
        )


def _get_namespaced(cls, resource_dict, api_value):
    errors = []
    for base in getattr(cls, "bases", []):
        api_group_name = _api_group_name(api_value=api_value)
        namespaced = base.id == "NamespacedResource"
        api_group = _api_group_dict(
            resource_dict=resource_dict, api_group_name=api_group_name
        )
        should_be_namespaced = api_group["namespaced"] == "true"

        if namespaced != should_be_namespaced:
            errors.append(
                f"Resource {cls.name} should be "
                f"{'namespaced' if should_be_namespaced else 'in cluster scope (not namespaced)'}"
            )
    return errors


def _get_api_group(api_value, cls, resource_dict):
    errors = []
    api_group_name = _api_group_name(api_value=api_value)

    if api_group_name not in resource_dict["api_group"]:
        errors.append(
            f"Resource {cls.name} api_group should be "
            f"{resource_dict['api_group']}. got {api_group_name}"
        )
    return errors


def _get_api_version(api_value, cls, resource_dict):
    errors = []
    api_group_name = _api_group_name(api_value=api_value)
    api_group = _api_group_dict(
        resource_dict=resource_dict, api_group_name=api_group_name
    )
    if api_value.lower() != api_group["api_version"]:
        desire_api_group = resource_dict["api_version"].split("/")[0]
        errors.append(
            f"Resource {cls.name} have api_version {api_value} "
            f"but should have api_group = {desire_api_group}"
        )

    return errors


def _resource_file():
    ocp_resources_exclude_files = ["resource.py", "utils.py", "__init__.py"]
    for root, dirs, files in os.walk("ocp_resources"):
        for _file in files:
            if _file in ocp_resources_exclude_files or not _file.endswith(".py"):
                continue
            yield os.path.join(root, _file)


@pytest.fixture()
def resources_definitions():
    file_ = (
        "https://raw.githubusercontent.com/RedHatQE/"
        "openshift-resources-definitions/main/resources_definitions.json"
    )
    content = requests.get(file_).content
    return json.loads(content)


@pytest.fixture()
def resources_definitions_errors(resources_definitions):
    errors = []
    for _file in _resource_file():
        with open(_file, "r") as fd:
            tree = ast.parse(source=fd.read())
            classes = [cls for cls in tree.body if isinstance(cls, ast.ClassDef)]
            for cls in classes:
                resource_dict = resources_definitions.get(cls.name)
                if not resource_dict:
                    continue

                bodies = [
                    body_
                    for body_ in getattr(cls, "body")
                    if isinstance(body_, ast.Assign)
                ]
                api_type, api_value = _get_api_group_and_version(bodies=bodies)
                errors.extend(
                    _get_namespaced(
                        cls=cls, resource_dict=resource_dict, api_value=api_value
                    )
                )
                errors.extend(
                    _process_api_type(
                        api_type=api_type,
                        api_value=api_value,
                        resource_dict=resource_dict,
                        cls=cls,
                    )
                )
    return errors


def test_resources_definitions(resources_definitions_errors):
    assert not resources_definitions_errors, "\n".join(resources_definitions_errors)



# -*- coding: utf-8 -*-

from openshift.dynamic.exceptions import ResourceNotFoundError

from ocp_resources.constants import PROTOCOL_ERROR_EXCEPTION_DICT, TIMEOUT_4MINUTES
from ocp_resources.logger import get_logger
from ocp_resources.resource import NamespacedResource
from ocp_resources.utils import TimeoutSampler
from ocp_resources.virtual_machine import VirtualMachine


LOGGER = get_logger(name=__name__)


class VirtualMachineRestore(NamespacedResource):
    """
    VirtualMachineRestore object.
    """

    api_group = NamespacedResource.ApiGroup.SNAPSHOT_KUBEVIRT_IO

    def __init__(
        self,
        name=None,
        namespace=None,
        vm_name=None,
        snapshot_name=None,
        client=None,
        teardown=True,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        **kwargs,
    ):
        super().__init__(
            name=name,
            namespace=namespace,
            client=client,
            teardown=teardown,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )
        self.vm_name = vm_name
        self.snapshot_name = snapshot_name

    def to_dict(self):
        res = super().to_dict()
        if self.yaml_file:
            return res

        spec = res.setdefault("spec", {})
        spec.setdefault("target", {})[
            "apiGroup"
        ] = NamespacedResource.ApiGroup.KUBEVIRT_IO
        spec["target"]["kind"] = VirtualMachine.kind
        spec["target"]["name"] = self.vm_name
        spec["virtualMachineSnapshotName"] = self.snapshot_name
        return res

    def wait_complete(self, status=True, timeout=TIMEOUT_4MINUTES):
        """
        Wait for VirtualMachineRestore to be in status complete

        Args:
            status: Expected status: True for a completed restore operation, False otherwise.
            timeout (int): Time to wait.

        Raises:
            TimeoutExpiredError: If timeout reached.
        """
        LOGGER.info(
            f"Wait for {self.kind} {self.name} status to be complete = {status}"
        )

        samples = TimeoutSampler(
            wait_timeout=timeout,
            sleep=1,
            exceptions_dict=PROTOCOL_ERROR_EXCEPTION_DICT,
            func=lambda: self.instance.get("status", {}).get("complete", None)
            == status,
        )
        for sample in samples:
            if sample:
                return

    def wait_restore_done(self, timeout=TIMEOUT_4MINUTES):
        """
        Wait for the the restore to be done. This check 2 parameters, the restore status to be complete
        and the VM status restoreInProgress to be None.

        Args:
            timeout (int): Time to wait.

        Raises:
            TimeoutExpiredError: If timeout reached.
        """
        self.wait_complete(timeout=timeout)

        vm = VirtualMachine(
            client=self.client,
            namespace=self.namespace,
            name=self.vm_name,
        )

        if vm.exists:
            return vm.wait_for_status_none(status="restoreInProgress", timeout=timeout)
        raise ResourceNotFoundError(f"VirtualMachine: {vm.name} not found")



from ocp_resources.resource import Resource


class MigrationPolicy(Resource):

    api_group = Resource.ApiGroup.MIGRATIONS_KUBEVIRT_IO

    def __init__(
        self,
        name=None,
        allow_auto_converge=None,
        allow_post_copy=None,
        bandwidth_per_migration=None,
        completion_timeout_per_gb=None,
        namespace_selector=None,
        vmi_selector=None,
        **kwargs,
    ):
        """
        Create MigrationPolicy object.

        Args:
            name (str): Migration Policy name
            allow_auto_converge (bool, optional)
            allow_post_copy (bool, optional)
            bandwidth_per_migration (str, optional, i.e. "1Gi")
            completion_timeout_per_gb (int, optional)
            namespace_selector (dict, optional): Dict of Namespace labels to match (e.g. {"project-owner": "redhat"})
            vmi_selector (dict, optional): Dict of VMI labels to match (e.g. {"best-vm": ""})
        """
        super().__init__(
            name=name,
            **kwargs,
        )
        self.allow_auto_converge = allow_auto_converge
        self.allow_post_copy = allow_post_copy
        self.bandwidth_per_migration = bandwidth_per_migration
        self.completion_timeout_per_gb = completion_timeout_per_gb
        self.namespace_selector = namespace_selector or {}
        self.vmi_selector = vmi_selector or {}

    def to_dict(self):
        res = super().to_dict()
        if self.yaml_file:
            return res
        spec = res.setdefault("spec", {})
        selectors = spec.setdefault("selectors", {})

        if self.allow_auto_converge is not None:
            res["spec"]["allowAutoConverge"] = self.allow_auto_converge
        if self.allow_post_copy is not None:
            res["spec"]["allowPostCopy"] = self.allow_post_copy
        if self.bandwidth_per_migration:
            res["spec"]["bandwidthPerMigration"] = self.bandwidth_per_migration
        if self.completion_timeout_per_gb:
            res["spec"]["completionTimeoutPerGiB"] = self.completion_timeout_per_gb

        if self.namespace_selector:
            selectors.setdefault("namespaceSelector", {}).setdefault(
                "matchLabels", self.namespace_selector
            )

        if self.vmi_selector:
            selectors.setdefault("virtualMachineInstanceSelector", {}).setdefault(
                "matchLabels", self.vmi_selector
            )

        return res



# -*- coding: utf-8 -*-

from ocp_resources.resource import NamespacedResource


class PeerAuthentication(NamespacedResource):
    """
    Peer Authentication object.
    """

    api_group = NamespacedResource.ApiGroup.SECURITY_ISTIO_IO

    class MtlsMode:
        """
        mTLS Traffic Mode object.
        """

        STRICT = "STRICT"



from ocp_resources.constants import TIMEOUT_4MINUTES
from ocp_resources.resource import NamespacedResource


class OperatorSource(NamespacedResource):
    api_group = NamespacedResource.ApiGroup.OPERATORS_COREOS_COM

    def __init__(
        self,
        name=None,
        namespace=None,
        registry_namespace=None,
        display_name=None,
        publisher=None,
        secret=None,
        client=None,
        teardown=True,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        **kwargs,
    ):
        super().__init__(
            name=name,
            namespace=namespace,
            client=client,
            teardown=teardown,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )
        self.registry_namespace = registry_namespace
        self.display_name = display_name
        self.publisher = publisher
        self.secret = secret

    def to_dict(self):
        res = super().to_dict()
        if self.yaml_file:
            return res

        res.update(
            {
                "spec": {
                    "type": "appregistry",
                    "endpoint": "https://quay.io/cnr",
                    "registryNamespace": self.registry_namespace,
                    "displayName": self.display_name,
                    "publisher": self.publisher,
                    "authorizationToken": {"secretName": self.secret},
                }
            }
        )

        return res



import logging
import os
from datetime import datetime
from logging.handlers import RotatingFileHandler

from colorlog import ColoredFormatter


LOGGER = logging.getLogger(__name__)


class DuplicateFilter(logging.Filter):
    def filter(self, record):
        repeated_number_exists = getattr(self, "repeated_number", None)
        current_log = (record.module, record.levelno, record.msg)
        if current_log != getattr(self, "last_log", None):
            self.last_log = current_log
            if repeated_number_exists:
                LOGGER.warning(f"Last log repeated {self.repeated_number} times.")

            self.repeated_number = 0
            return True
        if repeated_number_exists:
            self.repeated_number += 1
        else:
            self.repeated_number = 1
        return False


class WrapperLogFormatter(ColoredFormatter):
    def formatTime(self, record, datefmt=None):  # noqa: N802
        return datetime.fromtimestamp(record.created).isoformat()


def get_logger(name):
    log_level = os.environ.get("OPENSHIFT_PYTHON_WRAPPER_LOG_LEVEL", "INFO")
    log_file = os.environ.get("OPENSHIFT_PYTHON_WRAPPER_LOG_FILE", "")
    if log_level not in ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]:
        raise ValueError(f"Invalid log level: {log_level}")

    logger_obj = logging.getLogger(name)
    log_formatter = WrapperLogFormatter(
        fmt="%(asctime)s %(name)s %(log_color)s%(levelname)s%(reset)s %(message)s",
        log_colors={
            "DEBUG": "cyan",
            "INFO": "green",
            "WARNING": "yellow",
            "ERROR": "red",
            "CRITICAL": "red,bg_white",
        },
        secondary_log_colors={},
    )

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(fmt=log_formatter)
    console_handler.addFilter(filter=DuplicateFilter())

    logger_obj.addHandler(hdlr=console_handler)
    logger_obj.setLevel(level=log_level)
    logger_obj.addFilter(filter=DuplicateFilter())

    if log_file:
        log_handler = RotatingFileHandler(
            filename=log_file, maxBytes=100 * 1024 * 1024, backupCount=20
        )
        log_handler.setFormatter(fmt=log_formatter)
        log_handler.setLevel(level=log_level)
        logger_obj.addHandler(hdlr=log_handler)

    logger_obj.propagate = False
    return logger_obj



from ocp_resources.resource import Resource


class Infrastructure(Resource):
    """
    Infrastructure object.
    """

    api_group = Resource.ApiGroup.CONFIG_OPENSHIFT_IO

    class Type:
        BARE_METAL = "BareMetal"
        AWS = "AWS"
        OPENSTACK = "OpenStack"

    @property
    def platform(self):
        return self.instance.status.platformStatus.type



from ocp_resources.resource import NamespacedResource


class VirtualMachineInstancePreset(NamespacedResource):
    """
    VirtualMachineInstancePreset object.
    """

    api_group = NamespacedResource.ApiGroup.KUBEVIRT_IO



import shlex

import xmltodict
from openshift.dynamic.exceptions import ResourceNotFoundError

from ocp_resources.constants import PROTOCOL_ERROR_EXCEPTION_DICT, TIMEOUT_4MINUTES
from ocp_resources.logger import get_logger
from ocp_resources.node import Node
from ocp_resources.pod import Pod
from ocp_resources.resource import NamespacedResource
from ocp_resources.utils import TimeoutExpiredError, TimeoutSampler


LOGGER = get_logger(name=__name__)


class VirtualMachineInstance(NamespacedResource):
    """
    Virtual Machine Instance object, inherited from Resource.
    """

    api_group = NamespacedResource.ApiGroup.KUBEVIRT_IO

    class Status(NamespacedResource.Status):
        SCHEDULING = "Scheduling"

    def __init__(
        self,
        name=None,
        namespace=None,
        client=None,
        privileged_client=None,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        **kwargs,
    ):
        super().__init__(
            name=name,
            namespace=namespace,
            client=client,
            privileged_client=privileged_client,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )

    @property
    def _subresource_api_url(self):
        return (
            f"{self.client.configuration.host}/"
            f"apis/subresources.kubevirt.io/{self.api.api_version}/"
            f"namespaces/{self.namespace}/virtualmachineinstances/{self.name}"
        )

    def api_request(self, method, action, **params):
        return super().api_request(
            method=method, action=action, url=self._subresource_api_url, **params
        )

    def pause(self, timeout=TIMEOUT_4MINUTES, wait=False):
        self.api_request(method="PUT", action="pause")
        if wait:
            return self.wait_for_pause_status(pause=True, timeout=timeout)

    def unpause(self, timeout=TIMEOUT_4MINUTES, wait=False):
        self.api_request(method="PUT", action="unpause")
        if wait:
            return self.wait_for_pause_status(pause=False, timeout=timeout)

    @property
    def interfaces(self):
        return self.instance.status.interfaces

    @property
    def virt_launcher_pod(self):
        pods = list(
            Pod.get(
                dyn_client=self.privileged_client or self.client,
                namespace=self.namespace,
                label_selector=f"kubevirt.io=virt-launcher,kubevirt.io/created-by={self.instance.metadata.uid}",
            )
        )
        migration_state = self.instance.status.migrationState
        if migration_state:
            #  After VM migration there are two pods, one in Completed status and one in Running status.
            #  We need to return the Pod that is not in Completed status.
            for pod in pods:
                if migration_state.targetPod == pod.name:
                    return pod
        else:
            return pods[0]

        raise ResourceNotFoundError

    @property
    def virt_handler_pod(self):
        pods = list(
            Pod.get(
                dyn_client=self.privileged_client or self.client,
                label_selector="kubevirt.io=virt-handler",
            )
        )
        for pod in pods:
            if pod.instance["spec"]["nodeName"] == self.instance.status.nodeName:
                return pod

        raise ResourceNotFoundError

    def wait_until_running(self, timeout=TIMEOUT_4MINUTES, logs=True, stop_status=None):
        """
        Wait until VMI is running

        Args:
            timeout (int): Time to wait for VMI.
            logs (bool): True to extract logs from the VMI pod and from the VMI.
            stop_status (str): Status which should stop the wait and failed.

        Raises:
            TimeoutExpiredError: If VMI failed to run.
        """
        try:
            self.wait_for_status(
                status=self.Status.RUNNING, timeout=timeout, stop_status=stop_status
            )
        except TimeoutExpiredError:
            if not logs:
                raise

            virt_pod = self.virt_launcher_pod
            if virt_pod:
                LOGGER.debug(f"{virt_pod.name} *****LOGS*****")
                LOGGER.debug(virt_pod.log(container="compute"))

            raise

    def wait_for_pause_status(self, pause, timeout=TIMEOUT_4MINUTES):
        """
        Wait for Virtual Machine Instance to be paused / unpaused.
        Paused status is checked in libvirt and in the VMI conditions.

        Args:
            pause (bool): True for paused, False for unpause
            timeout (int): Time to wait for the resource.

        Raises:
            TimeoutExpiredError: If resource not exists.
        """
        LOGGER.info(
            f"Wait until {self.kind} {self.name} is "
            f"{'Paused' if pause else 'Unpuased'}"
        )
        self.wait_for_domstate_pause_status(pause=pause, timeout=timeout)
        self.wait_for_vmi_condition_pause_status(pause=pause, timeout=timeout)

    def wait_for_domstate_pause_status(self, pause, timeout=TIMEOUT_4MINUTES):
        pause_status = "paused" if pause else "running"
        samples = TimeoutSampler(
            wait_timeout=timeout,
            sleep=1,
            exceptions_dict=PROTOCOL_ERROR_EXCEPTION_DICT,
            func=self.get_domstate,
        )
        for sample in samples:
            if pause_status in sample:
                return

    def wait_for_vmi_condition_pause_status(self, pause, timeout=TIMEOUT_4MINUTES):
        samples = TimeoutSampler(
            wait_timeout=timeout,
            sleep=1,
            exceptions_dict=PROTOCOL_ERROR_EXCEPTION_DICT,
            func=self.get_vmi_active_condition,
        )
        for sample in samples:
            # VM in state change
            # We have commanded a [un]pause condition via the API but the CR has not been updated yet to match.
            # 'reason' may not exist yet
            # or
            # 'reason' may still exist after unpause if the CR has not been updated before we perform this check
            if (pause and not sample.get("reason")) or (
                sample.get("reason") == "PausedByUser" and not pause
            ):
                continue
            # Paused VM
            if pause and sample["reason"] == "PausedByUser":
                return
            # Unpaused VM
            if not (pause and sample.get("reason")):
                return

    @property
    def node(self):
        """
        Get the node name where the VM is running

        Returns:
            Node: Node
        """
        return Node(
            client=self.privileged_client or self.client,
            name=self.instance.status.nodeName,
        )

    def virsh_cmd(self, action):
        return shlex.split(
            f"virsh {self.virt_launcher_pod_hypervisor_connection_uri} {action} {self.namespace}_{self.name}"
        )

    def get_xml(self):
        """
        Get virtual machine instance XML

        Returns:
            xml_output(string): VMI XML in the multi-line string
        """
        return self.virt_launcher_pod.execute(
            command=self.virsh_cmd(action="dumpxml"),
            container="compute",
        )

    @property
    def virt_launcher_pod_user_uid(self):
        """
        Get Virt Launcher Pod User UID value

        Returns:
            Int: Virt Launcher Pod UID value
        """
        return self.virt_launcher_pod.instance.spec.securityContext.runAsUser

    @property
    def is_virt_launcher_pod_root(self):
        """
        Check if Virt Launcher Pod is Root

        Returns:
            Bool: True if Virt Launcher Pod is Root.
        """
        return not bool(self.virt_launcher_pod_user_uid)

    @property
    def virt_launcher_pod_hypervisor_connection_uri(self):
        """
        Get Virt Launcher Pod Hypervisor Connection URI

        Required to connect to Hypervisor for
        Non-Root Virt-Launcher Pod.

        Returns:
            String: Hypervisor Connection URI
        """
        return (
            ""
            if self.is_virt_launcher_pod_root
            else "-c qemu+unix:///session?socket=/var/run/libvirt/libvirt-sock"
        )

    def get_domstate(self):
        """
        Get virtual machine instance Status.

        Current workaround, as VM/VMI shows no status/phase == Paused yet.
        Bug: https://bugzilla.redhat.com/show_bug.cgi?id=1805178

        Returns:
            String: VMI Status as string
        """
        return self.virt_launcher_pod.execute(
            command=self.virsh_cmd(action="domstate"),
            container="compute",
        )

    def get_dommemstat(self):
        """
        Get virtual machine domain memory stats
        link: https://libvirt.org/manpages/virsh.html#dommemstat

        Returns:
            String: VMI domain memory stats as string
        """
        return self.virt_launcher_pod.execute(
            command=self.virsh_cmd(action="dommemstat"),
            container="compute",
        )

    def get_vmi_active_condition(self):
        """A VMI may have multiple conditions; the active one it the one with
        'lastTransitionTime'"""
        return {
            k: v
            for condition in self.instance.status.conditions
            for k, v in condition.items()
            if condition["lastTransitionTime"]
        }

    @property
    def xml_dict(self):
        """Get virtual machine instance XML as dict"""

        return xmltodict.parse(xml_input=self.get_xml(), process_namespaces=True)

    @property
    def guest_os_info(self):
        return self.api_request(method="GET", action="guestosinfo")

    @property
    def guest_fs_info(self):
        return self.api_request(method="GET", action="filesystemlist")

    @property
    def guest_user_info(self):
        return self.api_request(method="GET", action="userlist")

    @property
    def os_version(self):
        vmi_os_version = self.instance.status.guestOSInfo.get("version", {})
        if not vmi_os_version:
            LOGGER.warning(
                "Guest agent is not installed on the VM; OS version is not available."
            )
        return vmi_os_version

    def interface_ip(self, interface):
        iface_ip = [
            iface["ipAddress"]
            for iface in self.interfaces
            if iface["interfaceName"] == interface
        ]
        return iface_ip[0] if iface_ip else None



from ocp_resources.mtv import MTV
from ocp_resources.resource import NamespacedResource


class ForkliftController(NamespacedResource, MTV):
    """
    Migration Toolkit For Virtualization (MTV) ForkliftController Resource
    """

    api_group = NamespacedResource.ApiGroup.FORKLIFT_KONVEYOR_IO



import time

from openshift.dynamic.exceptions import ConflictError

from ocp_resources.constants import TIMEOUT_4MINUTES
from ocp_resources.logger import get_logger
from ocp_resources.resource import Resource
from ocp_resources.utils import TimeoutSampler


LOGGER = get_logger(name=__name__)

SLEEP = 1


class NodeNetworkState(Resource):
    api_group = Resource.ApiGroup.NMSTATE_IO

    def __init__(
        self,
        name=None,
        client=None,
        teardown=True,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        **kwargs,
    ):
        super().__init__(
            name=name,
            client=client,
            teardown=teardown,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )
        status = self.instance.to_dict()["status"]
        self.desired_state = status.get("desiredState", {"interfaces": []})

    def set_interface(self, interface):

        # First drop the interface is's already in the list
        interfaces = [
            iface
            for iface in self.desired_state["interfaces"]
            if iface["name"] != interface["name"]
        ]

        # Add the interface
        interfaces.append(interface)
        self.desired_state["interfaces"] = interfaces

    def to_dict(self):
        res = super().to_dict()
        if self.yaml_file:
            return res

        res.update(
            {
                "spec": {
                    "nodeName": self.name,
                    "managed": True,
                    "desiredState": self.desired_state,
                }
            }
        )
        return res

    def apply(self):
        resource = self.to_dict()
        retries_on_conflict = 3
        while True:
            try:
                resource["metadata"] = self.instance.to_dict()["metadata"]
                self.update(resource)
                break
            except ConflictError as e:
                retries_on_conflict -= 1
                if retries_on_conflict == 0:
                    raise e
                time.sleep(1)

    def wait_until_up(self, name):
        def _find_up_interface():
            iface = self.get_interface(name=name)
            if iface.get("state") == self.Interface.State.UP:
                return iface

            return None

        LOGGER.info(f"Checking if interface {name} is up -- {self.name}")
        samples = TimeoutSampler(
            wait_timeout=TIMEOUT_4MINUTES, sleep=SLEEP, func=_find_up_interface
        )
        for sample in samples:
            if sample:
                return

    def wait_until_deleted(self, name):
        LOGGER.info(f"Checking if interface {name} is deleted -- {self.name}")
        samples = TimeoutSampler(
            wait_timeout=self.delete_timeout,
            sleep=SLEEP,
            func=self.get_interface,
            name=name,
        )
        for sample in samples:
            if not sample:
                return

    @property
    def interfaces(self):
        return self.instance.to_dict()["status"]["currentState"]["interfaces"]

    @property
    def routes(self):
        return self.instance.status.currentState.routes

    def ipv4(self, iface):
        for interface in self.interfaces:
            if interface["name"] == iface:
                addresses = interface["ipv4"]["address"]
                if addresses:
                    return interface["ipv4"]["address"][0]["ip"]

    def get_interface(self, name):
        for interface in self.interfaces:
            if interface["name"] == name:
                return interface
        return {}



from ocp_resources.constants import TIMEOUT_4MINUTES
from ocp_resources.resource import NamespacedResource


class PodDisruptionBudget(NamespacedResource):
    """
    PodDisruptionBudget object
    """

    api_group = NamespacedResource.ApiGroup.POLICY

    def __init__(
        self,
        name=None,
        namespace=None,
        client=None,
        min_available=None,
        max_unavailable=None,
        selector=None,
        teardown=True,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        **kwargs,
    ):
        super().__init__(
            name=name,
            namespace=namespace,
            client=client,
            teardown=teardown,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )
        self.min_available = min_available
        self.max_unavailable = max_unavailable
        self.selector = selector

    def to_dict(self):
        res = super().to_dict()
        if self.yaml_file:
            return res

        update_dict = {
            "spec": {
                "selector": self.selector,
            },
        }

        if self.min_available is not None:
            update_dict["spec"]["minAvailable"] = self.min_available

        if self.max_unavailable is not None:
            update_dict["spec"]["maxUnavailable"] = self.max_unavailable

        res.update(update_dict)

        return res



from ocp_resources.resource import Resource


class ImageContentSourcePolicy(Resource):
    """
    ICSP object, inherited from Resource.
    """

    api_group = Resource.ApiGroup.OPERATOR_OPENSHIFT_IO



from ocp_resources.constants import TIMEOUT_4MINUTES
from ocp_resources.resource import NamespacedResource


class Machine(NamespacedResource):
    """
    Machine object.
    """

    api_group = NamespacedResource.ApiGroup.MACHINE_OPENSHIFT_IO

    def __init__(
        self,
        name=None,
        namespace=None,
        teardown=True,
        client=None,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        **kwargs,
    ):
        super().__init__(
            name=name,
            namespace=namespace,
            client=client,
            teardown=teardown,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )

    @property
    def cluster_name(self):
        return self.instance.metadata.labels[f"{self.api_group}/cluster-api-cluster"]

    @property
    def machine_role(self):
        return self.instance.metadata.labels[
            f"{self.api_group}/cluster-api-machine-role"
        ]

    @property
    def machine_type(self):
        return self.instance.metadata.labels[
            f"{self.api_group}/cluster-api-machine-type"
        ]

    @property
    def machineset_name(self):
        return self.instance.metadata.labels[f"{self.api_group}/cluster-api-machineset"]



import json

import kubernetes

from ocp_resources.constants import TIMEOUT_4MINUTES
from ocp_resources.logger import get_logger
from ocp_resources.node import Node
from ocp_resources.resource import NamespacedResource, kube_v1_api
from ocp_resources.utils import TimeoutWatch


LOGGER = get_logger(name=__name__)


class ExecOnPodError(Exception):
    def __init__(self, command, rc, out, err):
        self.cmd = command
        self.rc = rc
        self.out = out
        self.err = err

    def __str__(self):
        return (
            f"Command execution failure: "
            f"{self.cmd}, "
            f"RC: {self.rc}, "
            f"OUT: {self.out}, "
            f"ERR: {self.err}"
        )


class Pod(NamespacedResource):
    """
    Pod object, inherited from Resource.
    """

    api_version = NamespacedResource.ApiVersion.V1

    class Status(NamespacedResource.Status):
        CRASH_LOOPBACK_OFF = "CrashLoopBackOff"
        IMAGE_PULL_BACK_OFF = "ImagePullBackOff"
        ERR_IMAGE_PULL = "ErrImagePull"

    def __init__(
        self,
        name=None,
        namespace=None,
        client=None,
        teardown=True,
        privileged_client=None,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        **kwargs,
    ):
        super().__init__(
            name=name,
            namespace=namespace,
            client=client,
            teardown=teardown,
            privileged_client=privileged_client,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )

    @property
    def containers(self):
        """
        Get Pod containers

        Returns:
            list: List of Pod containers
        """
        return self.instance.spec.containers

    def execute(self, command, timeout=60, container=None, ignore_rc=False):
        """
        Run command on Pod

        Args:
            command (list): Command to run.
            timeout (int): Time to wait for the command.
            container (str): Container name where to exec the command.
            ignore_rc (bool): If True ignore error rc from the shell and return out.

        Returns:
            str: Command output.

        Raises:
            ExecOnPodError: If the command failed.
        """
        error_channel = {}
        stream_closed_error = "stream resp is closed"
        LOGGER.info(f"Execute {command} on {self.name} ({self.node.name})")
        resp = kubernetes.stream.stream(
            api_method=self._kube_v1_api.connect_get_namespaced_pod_exec,
            name=self.name,
            namespace=self.namespace,
            command=command,
            container=container or self.containers[0].name,
            stderr=True,
            stdin=False,
            stdout=True,
            tty=False,
            _preload_content=False,
        )

        timeout_watch = TimeoutWatch(timeout=timeout)
        while resp.is_open():
            resp.run_forever(timeout=2)
            try:
                error_channel = json.loads(
                    resp.read_channel(kubernetes.stream.ws_client.ERROR_CHANNEL)
                )
                break
            except json.decoder.JSONDecodeError:
                # Check remaining time, in order to throw exception
                # if remaining time reached zero
                if timeout_watch.remaining_time() <= 0:
                    raise ExecOnPodError(
                        command=command, rc=-1, out="", err=stream_closed_error
                    )

        rcstring = error_channel.get("status")
        if rcstring is None:
            raise ExecOnPodError(
                command=command, rc=-1, out="", err=stream_closed_error
            )

        stdout = resp.read_stdout(timeout=5)
        stderr = resp.read_stderr(timeout=5)

        if rcstring == "Success" or ignore_rc:
            return stdout

        if rcstring == "Failure":
            raise ExecOnPodError(command=command, rc=-1, out="", err=error_channel)

        returncode = [
            int(cause["message"])
            for cause in error_channel["details"]["causes"]
            if cause["reason"] == "ExitCode"
        ][0]

        raise ExecOnPodError(command=command, rc=returncode, out=stdout, err=stderr)

    def log(self, **kwargs):
        """
        Get Pod logs

        Returns:
            str: Pod logs.
        """
        return self._kube_v1_api.read_namespaced_pod_log(
            name=self.name, namespace=self.namespace, **kwargs
        )

    @property
    def node(self):
        """
        Get the node name where the Pod is running

        Returns:
            Node: Node
        """
        node_name = self.instance.spec.nodeName
        assert node_name, f"Node not found for pod {self.name}"
        return Node(
            client=self.privileged_client or self.client,
            name=node_name,
        )

    @property
    def ip(self):
        return self.instance.status.podIP

    @property
    def _kube_v1_api(self):
        return kube_v1_api(api_client=self.client.client)



from ocp_resources.resource import Resource


class MutatingWebhookConfiguration(Resource):
    """
    MutatingWebhookConfiguration object.
    """

    api_group = Resource.ApiGroup.ADMISSIONREGISTRATION_K8S_IO



from ocp_resources.constants import TIMEOUT_4MINUTES
from ocp_resources.resource import Resource


class NodeMaintenance(Resource):
    """
    Node Maintenance object, inherited from Resource.
    """

    api_group = Resource.ApiGroup.NODEMAINTENANCE_KUBEVIRT_IO

    def __init__(
        self,
        name=None,
        client=None,
        node=None,
        reason="TEST Reason",
        teardown=True,
        timeout=TIMEOUT_4MINUTES,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        **kwargs,
    ):
        super().__init__(
            name=name,
            client=client,
            teardown=teardown,
            timeout=timeout,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )
        self.node = node
        self.reason = reason

    def to_dict(self):
        res = super().to_dict()
        if self.yaml_file:
            return res

        assert self.node, "node is mandatory for create"
        res["spec"] = {"nodeName": self.node.name, "reason": self.reason}
        return res



from ocp_resources.constants import TIMEOUT_4MINUTES
from ocp_resources.resource import NamespacedResource


class SriovNetworkNodePolicy(NamespacedResource):
    """
    SriovNetworkNodePolicy object.
    """

    api_group = NamespacedResource.ApiGroup.SRIOVNETWORK_OPENSHIFT_IO

    def __init__(
        self,
        name=None,
        namespace=None,
        pf_names=None,
        root_devices=None,
        num_vfs=None,
        resource_name=None,
        client=None,
        priority=None,
        mtu=None,
        node_selector=None,
        teardown=True,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        **kwargs,
    ):
        super().__init__(
            name=name,
            namespace=namespace,
            client=client,
            teardown=teardown,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )
        self.pf_names = pf_names
        self.root_devices = root_devices
        self.num_vfs = num_vfs
        self.priority = priority
        self.resource_name = resource_name
        self.mtu = mtu
        self.node_selector = node_selector

    def to_dict(self):
        res = super().to_dict()
        if self.yaml_file:
            return res

        res["spec"] = {
            "deviceType": "vfio-pci",
            "nicSelector": {
                "pfNames": [self.pf_names],
                "rootDevices": [self.root_devices],
            },
            "numVfs": self.num_vfs,
            "resourceName": self.resource_name,
        }
        if self.mtu:
            res["spec"]["mtu"] = self.mtu
        if self.priority:
            res["spec"]["priority"] = self.priority
        if self.node_selector:
            res["spec"]["nodeSelector"] = self.node_selector
        else:
            res["spec"]["nodeSelector"] = {
                "feature.node.kubernetes.io/network-sriov.capable": "true"
            }
        return res



from ocp_resources.constants import TIMEOUT_4MINUTES
from ocp_resources.logger import get_logger
from ocp_resources.resource import NamespacedResource


LOGGER = get_logger(name=__name__)


class CatalogSource(NamespacedResource):
    api_group = NamespacedResource.ApiGroup.OPERATORS_COREOS_COM

    def __init__(
        self,
        name=None,
        namespace=None,
        client=None,
        source_type=None,
        image=None,
        display_name=None,
        publisher=None,
        teardown=True,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        update_strategy_registry_poll_interval=None,
        **kwargs,
    ):
        super().__init__(
            name=name,
            namespace=namespace,
            client=client,
            teardown=teardown,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )
        self.source_type = source_type
        self.image = image
        self.display_name = display_name
        self.publisher = publisher
        self.update_strategy_registry_poll_interval = (
            update_strategy_registry_poll_interval
        )

    def to_dict(self):
        res = super().to_dict()
        if self.yaml_file:
            return res

        res.update(
            {
                "spec": {
                    "sourceType": self.source_type,
                    "image": self.image,
                    "displayName": self.display_name,
                    "publisher": self.publisher,
                }
            }
        )

        if self.update_strategy_registry_poll_interval:
            res["spec"].update(
                {
                    "updateStrategy": {
                        "registryPoll": {
                            "interval": self.update_strategy_registry_poll_interval,
                        },
                    },
                }
            )

        return res



# -*- coding: utf-8 -*-

from ocp_resources.resource import NamespacedResource


class Gateway(NamespacedResource):
    """
    Gateway object.
    """

    api_group = NamespacedResource.ApiGroup.NETWORKING_ISTIO_IO



# -*- coding: utf-8 -*-

from ocp_resources.resource import NamespacedResource


class ChaosEngine(NamespacedResource):

    api_group = NamespacedResource.ApiGroup.LITMUS_IO

    class EngineStatus:
        INITIALIZED = "initialized"
        COMPLETED = "completed"
        STOPPED = "stopped"

    @property
    def engine_status(self):
        return self.instance.status["engineStatus"]

    @property
    def experiments_status(self):
        experiments = self.instance.status["experiments"]
        ret_value = {}
        for experiment in experiments:
            exp = {"verdict": experiment["verdict"], "status": experiment["status"]}
            ret_value[experiment["name"]] = exp
        return ret_value

    @property
    def success(self):
        exps = self.experiments_status
        return all(exp["verdict"] == "Pass" for exp in exps.values())



import re

from openshift.dynamic.exceptions import ConflictError

from ocp_resources.constants import TIMEOUT_4MINUTES
from ocp_resources.logger import get_logger
from ocp_resources.node import Node
from ocp_resources.node_network_configuration_enactment import (
    NodeNetworkConfigurationEnactment,
)
from ocp_resources.node_network_state import NodeNetworkState
from ocp_resources.resource import Resource, ResourceEditor
from ocp_resources.utils import TimeoutExpiredError, TimeoutSampler


LOGGER = get_logger(name=__name__)


class NNCPConfigurationFailed(Exception):
    pass


class NodeNetworkConfigurationPolicy(Resource):
    api_group = Resource.ApiGroup.NMSTATE_IO

    class Conditions:
        class Type:
            DEGRADED = "Degraded"
            AVAILABLE = "Available"

        class Reason:
            CONFIGURATION_PROGRESSING = "ConfigurationProgressing"
            SUCCESSFULLY_CONFIGURED = "SuccessfullyConfigured"
            FAILED_TO_CONFIGURE = "FailedToConfigure"
            NO_MATCHING_NODE = "NoMatchingNode"

    def __init__(
        self,
        name=None,
        client=None,
        capture=None,
        node_selector=None,
        node_selector_labels=None,
        teardown_absent_ifaces=True,
        teardown=True,
        mtu=None,
        ports=None,
        ipv4_enable=False,
        ipv4_dhcp=False,
        ipv4_auto_dns=True,
        ipv4_addresses=None,
        ipv6_enable=False,
        ipv6_dhcp=False,
        ipv6_auto_dns=True,
        ipv6_addresses=None,
        dns_resolver=None,
        routes=None,
        yaml_file=None,
        set_ipv4=True,
        set_ipv6=True,
        max_unavailable=None,
        state=None,
        success_timeout=480,
        delete_timeout=TIMEOUT_4MINUTES,
        **kwargs,
    ):
        """
        ipv4_addresses should be sent in this format:
        [{"ip": <ip1-string>, "prefix-length": <prefix-len1-int>},
         {"ip": <ip2-string>, "prefix-length": <prefix-len2-int>}, ...]
        For example:
        [{"ip": "10.1.2.3", "prefix-length": 24},
         {"ip": "10.4.5.6", "prefix-length": 24},
         {"ip": "10.7.8.9", "prefix-length": 23}]
        """
        super().__init__(
            name=name,
            client=client,
            teardown=teardown,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            node_selector=node_selector,
            node_selector_labels=node_selector_labels,
            **kwargs,
        )
        self.desired_state = {"interfaces": []}
        self.mtu = mtu
        self.capture = capture
        self.mtu_dict = {}
        self.ports = ports or []
        self.iface = None
        self.ipv4_enable = ipv4_enable
        self.ipv4_dhcp = ipv4_dhcp
        self.ipv4_auto_dns = ipv4_auto_dns
        self.ipv4_addresses = ipv4_addresses or []
        self.ipv4_iface_state = {}
        self.ipv6_enable = ipv6_enable
        self.ipv6_dhcp = ipv6_dhcp
        self.ipv6_autoconf = self.ipv6_dhcp
        self.ipv6_auto_dns = ipv6_auto_dns
        self.ipv6_addresses = ipv6_addresses
        self.dns_resolver = dns_resolver
        self.routes = routes
        self.state = state or self.Interface.State.UP
        self.set_ipv4 = set_ipv4
        self.set_ipv6 = set_ipv6
        self.success_timeout = success_timeout
        self.max_unavailable = max_unavailable
        self.ipv4_ports_backup_dict = {}
        self.ipv6_ports_backup_dict = {}
        self.nodes = self._nodes()
        self.teardown_absent_ifaces = teardown_absent_ifaces

    def _nodes(self):
        if self.node_selector:
            return list(Node.get(dyn_client=self.client, name=self.node_selector))
        if self.node_selector_labels:
            node_labels = ",".join(
                [
                    f"{label_key}={label_value}"
                    for label_key, label_value in self.node_selector_labels.items()
                ]
            )
            return list(Node.get(dyn_client=self.client, label_selector=node_labels))

    def set_interface(self, interface):
        if not self.res:
            self.res = super().to_dict()
        # First drop the interface if it's already in the list
        interfaces = [
            iface
            for iface in self.desired_state["interfaces"]
            if iface["name"] != interface["name"]
        ]
        # Add the interface
        interfaces.append(interface)
        self.desired_state["interfaces"] = interfaces
        self.res.setdefault("spec", {}).setdefault("desiredState", {})[
            "interfaces"
        ] = self.desired_state["interfaces"]

    def to_dict(self):
        self.res = super().to_dict()
        if self.yaml_file:
            return self.res

        if self.dns_resolver or self.routes or self.iface:
            self.res.setdefault("spec", {}).setdefault("desiredState", {})

        if self.node_selector_spec:
            self.res.setdefault("spec", {}).setdefault(
                "nodeSelector", self.node_selector_spec
            )

        if self.capture:
            self.res["spec"]["capture"] = self.capture

        if self.dns_resolver:
            self.res["spec"]["desiredState"]["dns-resolver"] = self.dns_resolver

        if self.routes:
            self.res["spec"]["desiredState"]["routes"] = self.routes

        if self.max_unavailable:
            self.res.setdefault("spec", {}).setdefault(
                "maxUnavailable", self.max_unavailable
            )

        if self.iface:
            """
            It's the responsibility of the caller to verify the desired configuration they send.
            For example: "ipv4.dhcp.enabled: false" without specifying any static IP address
            is a valid desired state and therefore not blocked in the code, but nmstate would
            reject it. Such configuration might be used for negative tests.
            """
            self.res = self.add_interface(
                iface=self.iface,
                state=self.state,
                set_ipv4=self.set_ipv4,
                ipv4_enable=self.ipv4_enable,
                ipv4_dhcp=self.ipv4_dhcp,
                ipv4_auto_dns=self.ipv4_auto_dns,
                ipv4_addresses=self.ipv4_addresses,
                set_ipv6=self.set_ipv6,
                ipv6_enable=self.ipv6_enable,
                ipv6_dhcp=self.ipv6_dhcp,
                ipv6_auto_dns=self.ipv6_auto_dns,
                ipv6_addresses=self.ipv6_addresses,
                ipv6_autoconf=self.ipv6_autoconf,
            )

        return self.res

    def add_interface(
        self,
        iface=None,
        name=None,
        type_=None,
        state=None,
        set_ipv4=True,
        ipv4_enable=False,
        ipv4_dhcp=False,
        ipv4_auto_dns=True,
        ipv4_addresses=None,
        set_ipv6=True,
        ipv6_enable=False,
        ipv6_dhcp=False,
        ipv6_auto_dns=True,
        ipv6_addresses=None,
        ipv6_autoconf=False,
    ):
        #  If self.res is already defined (from to_dict()), don't call it again.
        if not self.res:
            self.res = self.to_dict()

        self.res.setdefault("spec", {}).setdefault("desiredState", {})
        if not iface:
            iface = {
                "name": name,
                "type": type_,
                "state": state,
            }
        if set_ipv4:
            if isinstance(set_ipv4, str):
                iface["ipv4"] = set_ipv4

            else:
                iface["ipv4"] = {
                    "enabled": ipv4_enable,
                    "dhcp": ipv4_dhcp,
                    "auto-dns": ipv4_auto_dns,
                }
                if ipv4_addresses:
                    iface["ipv4"]["address"] = ipv4_addresses

        if set_ipv6:
            if isinstance(set_ipv6, str):
                iface["ipv6"] = set_ipv6

            else:
                iface["ipv6"] = {
                    "enabled": ipv6_enable,
                    "dhcp": ipv6_dhcp,
                    "auto-dns": ipv6_auto_dns,
                    "autoconf": ipv6_autoconf,
                }
                if ipv6_addresses:
                    iface["ipv6"]["address"] = ipv6_addresses

        self.set_interface(interface=iface)
        return self.res

    def _get_port_from_nns(self, port_name):
        if not self.nodes:
            return None

        nns = NodeNetworkState(name=self.nodes[0].name)
        _port = [_iface for _iface in nns.interfaces if _iface["name"] == port_name]
        return _port[0] if _port else None

    def _ports_backup(self, ip_family):
        for port in self.ports:
            _port = self._get_port_from_nns(port_name=port)
            if _port:
                self.ipv4_ports_backup_dict[port] = _port[ip_family]

    def ipv4_ports_backup(self):
        self._ports_backup(ip_family="ipv4")

    def ipv6_ports_backup(self):
        self._ports_backup(ip_family="ipv6")

    def add_ports(self):
        for port in self.ports:
            _port = self._get_port_from_nns(port_name=port)
            if _port:
                ipv4_backup = self.ipv4_ports_backup_dict.get(port)
                ipv6_backup = self.ipv6_ports_backup_dict.get(port)
                if ipv4_backup or ipv6_backup:
                    iface = {
                        "name": port,
                        "type": _port["type"],
                        "state": _port["state"],
                    }
                    if ipv4_backup:
                        iface["ipv4"] = ipv4_backup

                    if ipv6_backup:
                        iface["ipv6"] = ipv6_backup

                    self.set_interface(interface=iface)

    def apply(self, resource=None):
        resource = resource if resource else super().to_dict()
        samples = TimeoutSampler(
            wait_timeout=3,
            sleep=1,
            exceptions_dict={ConflictError: []},
            func=self.update,
            resource_dict=resource,
        )
        LOGGER.info(f"Applying {resource}")
        for _ in samples:
            return

    def deploy(self, wait=False):
        self.ipv4_ports_backup()
        self.ipv6_ports_backup()

        self.create(body=self.res, wait=wait)
        try:
            self.wait_for_status_success()
            return self
        except Exception as exp:
            LOGGER.error(exp)
            super().__exit__(exception_type=None, exception_value=None, traceback=None)
            raise

    def clean_up(self):
        if self.teardown_absent_ifaces:
            try:
                self._absent_interface()
                self.wait_for_status_success()
            except Exception as exp:
                LOGGER.error(exp)

        super().clean_up()

    def _absent_interface(self):
        for _iface in self.desired_state["interfaces"]:
            _iface["state"] = self.Interface.State.ABSENT
            self.set_interface(interface=_iface)

        if self.ports:
            self.add_ports()

        ResourceEditor(
            patches={
                self: {
                    "spec": {
                        "desiredState": {"interfaces": self.desired_state["interfaces"]}
                    }
                }
            }
        ).update()

    @property
    def status(self):
        for condition in self.instance.status.conditions:
            if condition["type"] == self.Conditions.Type.AVAILABLE:
                return condition["reason"]

    def wait_for_configuration_conditions_unknown_or_progressing(self, wait_timeout=30):
        samples = TimeoutSampler(
            wait_timeout=wait_timeout,
            sleep=1,
            func=lambda: self.instance.status.conditions,
        )
        for sample in samples:
            if (
                sample
                and sample[0]["type"] == self.Conditions.Type.AVAILABLE
                and (
                    sample[0]["status"] == self.Condition.Status.UNKNOWN
                    or sample[0]["reason"]
                    == self.Conditions.Reason.CONFIGURATION_PROGRESSING
                )
            ):
                return sample

    def _process_failed_status(self, failed_condition_reason):
        last_err_msg = None
        for failed_nnce in self._get_failed_nnce():
            nnce_name = failed_nnce.instance.metadata.name
            nnce_dict = failed_nnce.instance.to_dict()
            for cond in nnce_dict["status"]["conditions"]:
                err_msg = self._get_nnce_error_msg(
                    nnce_name=nnce_name, nnce_condition=cond
                )
                if err_msg:
                    last_err_msg = err_msg

        raise NNCPConfigurationFailed(
            f"Reason: {failed_condition_reason}\n{last_err_msg}"
        )

    def wait_for_status_success(self):
        failed_condition_reason = self.Conditions.Reason.FAILED_TO_CONFIGURE
        no_match_node_condition_reason = self.Conditions.Reason.NO_MATCHING_NODE

        # if we get here too fast there are no conditions, we need to wait.
        self.wait_for_configuration_conditions_unknown_or_progressing()

        samples = TimeoutSampler(
            wait_timeout=self.success_timeout, sleep=1, func=lambda: self.status
        )
        try:
            for sample in samples:
                if sample == self.Conditions.Reason.SUCCESSFULLY_CONFIGURED:
                    LOGGER.info(f"NNCP {self.name} configured Successfully")
                    return sample

                elif sample == no_match_node_condition_reason:
                    raise NNCPConfigurationFailed(
                        f"{self.name}. Reason: {no_match_node_condition_reason}"
                    )

                elif sample == failed_condition_reason:
                    self._process_failed_status(
                        failed_condition_reason=failed_condition_reason
                    )

        except (TimeoutExpiredError, NNCPConfigurationFailed):
            LOGGER.error(
                f"Unable to configure NNCP {self.name} "
                f"{f'nodes: {[node.name for node in self.nodes]}' if self.nodes else ''}"
            )
            raise

    @property
    def nnces(self):
        nnces = []
        for nnce in NodeNetworkConfigurationEnactment.get(dyn_client=self.client):
            if nnce.name.endswith(f".{self.name}"):
                nnces.append(nnce)
        return nnces

    def node_nnce(self, node_name):
        nnce = [
            nnce for nnce in self.nnces if nnce.labels["nmstate.io/node"] == node_name
        ]
        return nnce[0] if nnce else None

    @staticmethod
    def _get_nnce_error_msg(nnce_name, nnce_condition):
        err_msg = ""
        nnce_prefix = f"NNCE {nnce_name}"
        nnce_msg = nnce_condition.get("message")
        if not nnce_msg:
            return err_msg

        errors = nnce_msg.split("->")
        if errors:
            err_msg += f"{nnce_prefix}: {errors[0]}"
            if len(errors) > 1:
                err_msg += errors[-1]

        libnmstate_err = re.findall(r"libnmstate.error.*", nnce_msg)
        if libnmstate_err:
            err_msg += f"{nnce_prefix}: {libnmstate_err[0]}"

        return err_msg

    def _get_failed_nnce(self):
        for nnce in self.nnces:
            try:
                nnce.wait_for_conditions()
            except TimeoutExpiredError:
                LOGGER.error(f"Failed to get NNCE {nnce.name} status")
                continue

            for nnce_cond in nnce.instance.status.conditions:
                if (
                    nnce_cond.type == "Failing"
                    and nnce_cond.status == Resource.Condition.Status.TRUE
                ):
                    yield nnce



from ocp_resources.constants import TIMEOUT_4MINUTES
from ocp_resources.resource import Resource


class Project(Resource):
    """
    Project object.
    This is openshift's object which represents Namespace
    """

    api_group = Resource.ApiGroup.PROJECT_OPENSHIFT_IO

    class Status(Resource.Status):
        ACTIVE = "Active"

    def clean_up(self):
        Project(name=self.name).delete(wait=True)


class ProjectRequest(Resource):
    """
    RequestProject object.
    Resource which adds Project and grand
    full access to user who originated this request
    """

    api_group = Resource.ApiGroup.PROJECT_OPENSHIFT_IO

    def __init__(
        self,
        name=None,
        client=None,
        teardown=True,
        timeout=TIMEOUT_4MINUTES,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        **kwargs,
    ):
        super().__init__(
            name=name,
            client=client,
            teardown=teardown,
            timeout=timeout,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )

    def clean_up(self):
        Project(name=self.name).delete(wait=True)

    def client_wait_deleted(self, timeout):
        """
        client-side Wait until resource is deleted

        Args:
            timeout (int): Time to wait for the resource.

        """
        super().client_wait_deleted(timeout=timeout)



# -*- coding: utf-8 -*-
from ocp_resources.constants import TIMEOUT_4MINUTES
from ocp_resources.logger import get_logger
from ocp_resources.resource import NamespacedResource


LOGGER = get_logger(name=__name__)


class Route(NamespacedResource):
    """
    OpenShift Route object.
    """

    api_group = NamespacedResource.ApiGroup.ROUTE_OPENSHIFT_IO

    def __init__(
        self,
        name=None,
        namespace=None,
        client=None,
        service=None,
        destination_ca_cert=None,
        teardown=True,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        **kwargs,
    ):
        super().__init__(
            name=name,
            namespace=namespace,
            client=client,
            teardown=teardown,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )
        self.service = service
        self.destination_ca_cert = destination_ca_cert

    def to_dict(self):
        res = super().to_dict()
        if self.yaml_file:
            return res

        if self.service:
            res.update({"spec": {"to": {"kind": "Service", "name": self.service}}})
        if self.destination_ca_cert:
            res["spec"]["tls"] = {
                "destinationCACertificate": self.destination_ca_cert,
                "termination": "reencrypt",
            }
        return res

    @property
    def exposed_service(self):
        """
        returns the service the route is exposing
        """
        return self.instance.spec.to.name

    @property
    def host(self):
        """
        returns hostname that is exposing the service
        """
        return self.instance.spec.host

    @property
    def ca_cert(self):
        """
        returns destinationCACertificate
        """
        return self.instance.spec.tls.destinationCACertificate

    @property
    def termination(self):
        """
        returns a secured route using re-encrypt termination
        """
        return self.instance.spec.tls.termination



from ocp_resources.constants import TIMEOUT_4MINUTES
from ocp_resources.logger import get_logger
from ocp_resources.resource import NamespacedResource
from ocp_resources.utils import TimeoutExpiredError, TimeoutSampler


LOGGER = get_logger(name=__name__)


TIMEOUT_5MINUTES = 300


class MachineSet(NamespacedResource):
    """
    Machineset object.

    Args:
        cluster_name (str): OpenShift cluster name.
        machine_role (str): machine role. e.g.: 'worker'.
        machine_type (str): machine role. e.g.: 'worker'.
        replicas (int): amount of replicas the machine-set will have.
        provider_spec (dict): provider spec information.

        provider spec example:

    {
        "value": {
            "apiVersion": "ovirtproviderconfig.machine.openshift.io/v1beta1",
            "auto_pinning_policy": "none",
            "cluster_id": "5612af70-f4f5-455d-b7df-fbad66accc38",
            "cpu": {
                "cores": 8,
                "sockets": 1,
                "threads": 1
            },
            "credentialsSecret": {
                "name": "ovirt-credentials"
            },
            "kind": "OvirtMachineProviderSpec",
            "memory_mb": 16000,
            "os_disk": {
                "size_gb": 31
            },
            "template_name": "ge2n1-gcwmg-rhcos",
            "type": "server",
            "userDataSecret": {
                "name": "worker-user-data"
            }
        }
    }
    """

    api_group = NamespacedResource.ApiGroup.MACHINE_OPENSHIFT_IO

    def __init__(
        self,
        cluster_name=None,
        name=None,
        namespace=None,
        teardown=True,
        client=None,
        machine_role="worker",
        machine_type="worker",
        replicas=1,
        provider_spec=None,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        **kwargs,
    ):
        super().__init__(
            name=name,
            namespace=namespace,
            client=client,
            teardown=teardown,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )
        self.replicas = replicas
        self.cluster_name = cluster_name
        self.machine_role = machine_role
        self.machine_type = machine_type
        self.provider_spec = provider_spec or {}

    def to_dict(self):
        res = super().to_dict()
        if self.yaml_file:
            return res

        _spec, _metadata, _labels = ("spec", "metadata", "labels")
        (
            _cluster_api_cluster,
            _cluster_api_machine_role,
            _cluster_api_machine_type,
            _cluster_api_machineset,
        ) = (
            "cluster-api-cluster",
            "cluster-api-machine-role",
            "cluster-api-machine-type",
            "cluster-api-machineset",
        )

        res[_metadata][_labels] = {
            f"{self.api_group}/{_cluster_api_cluster}": self.cluster_name,
            f"{self.api_group}/{_cluster_api_machine_role}": self.machine_role,
            f"{self.api_group}/{_cluster_api_machine_type}": self.machine_type,
        }

        res[_spec] = {
            "replicas": self.replicas,
            "selector": {
                "matchLabels": {
                    f"{self.api_group}/{_cluster_api_cluster}": self.cluster_name,
                    f"{self.api_group}/{_cluster_api_machineset}": f"{self.cluster_name}-{self.machine_role}",
                }
            },
            "template": {
                _metadata: {
                    _labels: {
                        f"{self.api_group}/{_cluster_api_cluster}": self.cluster_name,
                        f"{self.api_group}/{_cluster_api_machine_role}": self.machine_role,
                        f"{self.api_group}/{_cluster_api_machine_type}": self.machine_type,
                        f"{self.api_group}/{_cluster_api_machineset}": f"{self.cluster_name}-{self.machine_role}",
                    }
                },
                _spec: {"providerSpec": self.provider_spec},
            },
        }
        return res

    @property
    def available_replicas(self):
        return self.instance.status.availableReplicas

    @property
    def ready_replicas(self):
        return self.instance.status.readyReplicas

    @property
    def desired_replicas(self):
        return self.instance.status.replicas

    @property
    def provider_spec_value(self):
        return self.instance.spec.template.spec.providerSpec.value

    def wait_for_replicas(self, timeout=TIMEOUT_5MINUTES, sleep=1):
        """
        Wait for machine-set replicas to reach 'ready' state.

        Args:
            timeout (int): maximum time to wait_for_replicas for the 'ready' state.
            sleep (int): sleep time between each sample.

        Returns:
            bool: True if machine-set reached 'ready' state, False otherwise.
        """
        try:
            return any(
                ready_replicas and ready_replicas == self.desired_replicas
                for ready_replicas in TimeoutSampler(
                    wait_timeout=timeout, sleep=sleep, func=lambda: self.ready_replicas
                )
            )
        except TimeoutExpiredError:
            LOGGER.error(
                f"Machine-set {self.name} replicas failed to reach into 'ready' state, actual ready replicas: "
                f"{self.ready_replicas}, desired replicas: {self.desired_replicas}"
            )
            return False

    def scale_replicas(
        self, replicas, wait_timeout=TIMEOUT_5MINUTES, sleep=1, wait=True
    ):
        """
        Scale down/up a machine-set replicas.

        Args:
            replicas (int): num of replicas to scale_replicas to.
            wait_timeout (int): maximum time to wait_for_replicas for scaling the machine-set.
            sleep (int): sleep time between each sample of the machine-set state.
            wait (bool): True if waiting for machine-set to reach into 'ready' state, False otherwise.

        Returns:
            bool: True if scaling the machine-set was successful or wait=False, False otherwise.
        """
        body = super().to_dict()
        body.update({"spec": {"replicas": replicas}})

        LOGGER.info(
            f"Scale machine-set from {self.desired_replicas} replicas to {replicas} replicas"
        )
        self.update(resource_dict=body)
        if wait:
            return self.wait_for_replicas(timeout=wait_timeout, sleep=sleep)
        return True



from ocp_resources.constants import TIMEOUT_4MINUTES
from ocp_resources.resource import NamespacedResource


class Job(NamespacedResource):
    """
    Job object.
    """

    api_group = NamespacedResource.ApiGroup.BATCH

    class Condition(NamespacedResource.Condition):
        COMPLETE = "Complete"

    def __init__(
        self,
        name=None,
        namespace=None,
        client=None,
        teardown=True,
        privileged_client=None,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        backoff_limit=None,
        restart_policy=None,
        service_account=None,
        containers=None,
        **kwargs,
    ):
        super().__init__(
            name=name,
            namespace=namespace,
            client=client,
            teardown=teardown,
            privileged_client=privileged_client,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )
        self.restart_policy = restart_policy
        self.backoff_limit = backoff_limit
        self.service_account = service_account
        self.containers = containers

    def to_dict(self):
        self.res = super().to_dict()
        self.res.setdefault("spec", {})

        if self.backoff_limit is not None:
            self.res["spec"]["backoffLimit"] = self.backoff_limit

        if self.containers:
            self.res["spec"].setdefault("template", {}).setdefault("spec", {})
            self.res["spec"]["template"]["spec"]["containers"] = self.containers

            if self.service_account:
                self.res["spec"]["template"]["spec"][
                    "serviceAccount"
                ] = self.service_account

            if self.restart_policy:
                self.res["spec"]["template"]["spec"][
                    "restartPolicy"
                ] = self.restart_policy

        return self.res



from ocp_resources.resource import Resource


class MachineConfigPool(Resource):
    api_group = Resource.ApiGroup.MACHINECONFIGURATION_OPENSHIFT_IO

    class Status(Resource.Status):
        UPDATED = "Updated"
        UPDATING = "Updating"



from ocp_resources.constants import TIMEOUT_4MINUTES
from ocp_resources.mtv import MTV
from ocp_resources.resource import NamespacedResource


class Host(NamespacedResource, MTV):
    """
    Migration Toolkit For Virtualization (MTV) Host resource.
    """

    api_group = NamespacedResource.ApiGroup.FORKLIFT_KONVEYOR_IO

    def __init__(
        self,
        name=None,
        namespace=None,
        host_id=None,
        ip_address=None,
        provider_name=None,
        provider_namespace=None,
        secret_name=None,
        secret_namespace=None,
        client=None,
        teardown=True,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        **kwargs,
    ):
        super().__init__(
            name=name,
            namespace=namespace,
            client=client,
            teardown=teardown,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )
        self.host_id = host_id
        self.ip_address = ip_address
        self.provider_name = provider_name
        self.provider_namespace = provider_namespace
        self.secret_name = secret_name or f"{self.name}-secret"
        self.secret_namespace = secret_namespace or self.namespace
        self.condition_message_ready = self.ConditionMessage.HOST_READY

    def to_dict(self):
        res = super().to_dict()
        if self.yaml_file:
            return res

        res.update(
            {
                "spec": {
                    "id": self.host_id,
                    "ipAddress": self.ip_address,
                    "secret": {
                        "name": self.secret_name,
                        "namespace": self.secret_namespace,
                    },
                    "provider": {
                        "name": self.provider_name,
                        "namespace": self.provider_namespace,
                    },
                }
            }
        )

        return res



# -*- coding: utf-8 -*-

from ocp_resources.resource import NamespacedResource


class Restore(NamespacedResource):
    """
    Restore object.
    """

    api_group = NamespacedResource.ApiGroup.VELERO_IO

    def __init__(
        self,
        name=None,
        namespace=None,
        included_namespaces=None,
        backup_name=None,
        client=None,
        teardown=False,
        privileged_client=None,
        yaml_file=None,
        **kwargs,
    ):
        if not backup_name:
            raise ValueError("backup_name can't be None")

        super().__init__(
            name=name,
            namespace=namespace,
            client=client,
            teardown=teardown,
            privileged_client=privileged_client,
            yaml_file=yaml_file,
            **kwargs,
        )

        self.included_namespaces = included_namespaces
        self.backup_name = backup_name

    def to_dict(self):
        res = super().to_dict()
        if self.yaml_file:
            return res
        res.update(
            {
                "spec": {
                    "backupName": self.backup_name,
                }
            }
        )
        if self.included_namespaces:
            res["spec"]["includedNamespaces"] = self.included_namespaces
        return res



from ocp_resources.constants import TIMEOUT_4MINUTES
from ocp_resources.resource import NamespacedResource


class DataImportCron(NamespacedResource):
    """
    DataImportCron object.
    """

    api_group = NamespacedResource.ApiGroup.CDI_KUBEVIRT_IO

    def __init__(
        self,
        name=None,
        namespace=None,
        client=None,
        image_stream=None,
        url=None,
        cert_configmap=None,
        pull_method=None,
        storage_class=None,
        size=None,
        schedule=None,
        garbage_collect=None,
        managed_data_source=None,
        imports_to_keep=None,
        bind_immediate_annotation=None,
        teardown=True,
        privileged_client=None,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        **kwargs,
    ):
        super().__init__(
            name=name,
            namespace=namespace,
            client=client,
            teardown=teardown,
            privileged_client=privileged_client,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )
        self.image_stream = image_stream
        self.url = url
        self.cert_configmap = cert_configmap
        self.pull_method = pull_method
        self.storage_class = storage_class
        self.size = size
        self.schedule = schedule
        self.garbage_collect = garbage_collect
        self.managed_data_source = managed_data_source
        self.imports_to_keep = imports_to_keep
        self.bind_immediate_annotation = bind_immediate_annotation

    def to_dict(self):
        res = super().to_dict()
        if self.yaml_file:
            return res
        if self.image_stream and self.url:
            raise ValueError("imageStream and url can not coexist")
        res.update(
            {
                "spec": {
                    "template": {
                        "spec": {
                            "source": {"registry": {"pullMethod": self.pull_method}},
                            "storage": {
                                "resources": {"requests": {"storage": self.size}}
                            },
                        }
                    }
                }
            }
        )
        spec = res["spec"]["template"]["spec"]
        if self.bind_immediate_annotation:
            res["metadata"].setdefault("annotations", {}).update(
                {
                    f"{NamespacedResource.ApiGroup.CDI_KUBEVIRT_IO}/storage.bind.immediate.requested": "true"
                }
            )
        if self.image_stream:
            spec["source"]["registry"]["imageStream"] = self.image_stream
        if self.url:
            spec["source"]["registry"]["url"] = self.url
        if self.cert_configmap:
            spec["source"]["registry"]["certConfigMap"] = self.cert_configmap
        if self.storage_class:
            spec["storage"]["storageClassName"] = self.storage_class
        if self.schedule:
            res["spec"]["schedule"] = self.schedule
        if self.garbage_collect:
            res["spec"]["garbageCollect"] = self.garbage_collect
        if self.managed_data_source:
            res["spec"]["managedDataSource"] = self.managed_data_source
        if self.imports_to_keep:
            res["spec"]["importsToKeep"] = self.imports_to_keep

        return res



# -*- coding: utf-8 -*-
from ocp_resources.constants import TIMEOUT_4MINUTES
from ocp_resources.resource import Resource


class ClusterRole(Resource):
    """
    ClusterRole object
    """

    api_group = Resource.ApiGroup.RBAC_AUTHORIZATION_K8S_IO

    def __init__(
        self,
        name=None,
        client=None,
        api_groups=None,
        permissions_to_resources=None,
        verbs=None,
        teardown=True,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        **kwargs,
    ):
        super().__init__(
            client=client,
            name=name,
            teardown=teardown,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )
        self.api_groups = api_groups
        self.permissions_to_resources = permissions_to_resources
        self.verbs = verbs
        self.desired_state = {"rules": []}

    def to_dict(self):
        self.res = super().to_dict()
        if self.yaml_file:
            return self.res

        if self.permissions_to_resources:
            self.add_rule(
                api_groups=self.api_groups,
                permissions_to_resources=self.permissions_to_resources,
                verbs=self.verbs,
            )

        return self.res

    def add_rule(
        self,
        api_groups=None,
        permissions_to_resources=None,
        verbs=None,
    ):
        if not self.res:
            self.res = super().to_dict()

        rule = {}
        if api_groups:
            rule["apiGroups"] = api_groups
        if permissions_to_resources:
            rule["resources"] = permissions_to_resources
        if verbs:
            rule["verbs"] = verbs
        if rule:
            self._set_rule(rule=rule)

        return self.res

    def _set_rule(self, rule):

        self.desired_state["rules"].append(rule)
        self.res["rules"] = self.desired_state["rules"]



# -*- coding: utf-8 -*-

from ocp_resources.resource import NamespacedResource


class ChaosResult(NamespacedResource):
    """
    ChaosResult resource used in Litmus experiments.
    """

    api_group = NamespacedResource.ApiGroup.LITMUS_IO



from ocp_resources.resource import Resource


class ConsoleCLIDownload(Resource):
    """
    ConsoleCLIDownload object, inherited from Resource.
    """

    api_group = Resource.ApiGroup.CONSOLE_OPENSHIFT_IO



from ocp_resources.resource import Resource


class Node(Resource):
    """
    Node object, inherited from Resource.
    """

    api_version = Resource.ApiVersion.V1

    class Status(Resource.Status):
        READY = "Ready"
        SCHEDULING_DISABLED = "Ready,SchedulingDisabled"

    @property
    def kubelet_ready(self):
        return any(
            [
                stat
                for stat in self.instance.status.conditions
                if stat["reason"] == "KubeletReady"
                and stat["status"] == self.Condition.Status.TRUE
            ]
        )

    @property
    def machine_name(self):
        return self.instance.metadata.annotations[
            f"{self.ApiGroup.MACHINE_OPENSHIFT_IO}/machine"
        ].split("/")[-1]

    @property
    def internal_ip(self):
        for addr in self.instance.status.addresses:
            if addr.type == "InternalIP":
                return addr.address

    @property
    def hostname(self):
        return self.labels["kubernetes.io/hostname"]

    @property
    def taints(self):
        return self.instance.get("spec", {}).get("taints")



# -*- coding: utf-8 -*-

from ocp_resources.resource import NamespacedResource


class ServiceMeshMemberRoll(NamespacedResource):
    """
    Service Mesh Member Roll object.
    """

    api_group = NamespacedResource.ApiGroup.MAISTRA_IO

    class Status:
        CONFIGURED = "Configured"



from ocp_resources.resource import NamespacedResource


class PrometheusRule(NamespacedResource):
    """
    Prometheus Rule object.
    """

    api_group = NamespacedResource.ApiGroup.MONITORING_COREOS_COM



from ocp_resources.resource import Resource


class NodeNetworkConfigurationEnactment(Resource):
    api_group = Resource.ApiGroup.NMSTATE_IO

    class Conditions:
        class Type:
            FAILING = "Failing"
            AVAILABLE = "Available"
            PROGRESSING = "Progressing"
            PENDING = "Pending"
            ABORTED = "Aborted"

        class Reason:
            CONFIGURATION_PROGRESSING = "ConfigurationProgressing"
            SUCCESSFULLY_CONFIGURED = "SuccessfullyConfigured"
            FAILED_TO_CONFIGURE = "FailedToConfigure"
            CONFIGURATION_ABORTED = "ConfigurationAborted"
            MAX_UNAVAILABLE_LIMIT_REACHED = "MaxUnavailableLimitReached"



from ocp_resources.constants import TIMEOUT_4MINUTES
from ocp_resources.resource import NamespacedResource


class VirtualMachineInstanceMigration(NamespacedResource):
    api_group = NamespacedResource.ApiGroup.KUBEVIRT_IO

    def __init__(
        self,
        name=None,
        namespace=None,
        vmi=None,
        client=None,
        teardown=True,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        **kwargs,
    ):
        super().__init__(
            name=name,
            namespace=namespace,
            client=client,
            teardown=teardown,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )
        self._vmi = vmi

    def to_dict(self):
        # When creating VirtualMachineInstanceMigration vmi is mandatory but when calling get()
        # we cannot pass vmi.
        res = super().to_dict()
        if self.yaml_file:
            return res

        assert self._vmi, "vmi is mandatory for create"
        res["spec"] = {"vmiName": self._vmi.name}
        return res



from openshift.dynamic.exceptions import NotFoundError
from urllib3.exceptions import ProtocolError


PROTOCOL_ERROR_EXCEPTION_DICT = {ProtocolError: []}
NOT_FOUND_ERROR_EXCEPTION_DICT = {NotFoundError: []}
TIMEOUT_4MINUTES = 240



from ocp_resources.constants import TIMEOUT_4MINUTES
from ocp_resources.mtv import MTV
from ocp_resources.resource import NamespacedResource


class Hook(NamespacedResource, MTV):
    """
    Migration Tool for Virtualization (MTV) Plan's Hook Resource.
    """

    api_group = NamespacedResource.ApiGroup.FORKLIFT_KONVEYOR_IO

    def __init__(
        self,
        name=None,
        namespace=None,
        image="quay.io/konveyor/hook-runner:latest",
        playbook=None,
        client=None,
        teardown=True,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        **kwargs,
    ):
        """
        Args:
            image (str): Path to an ansible image
            playbook (str): Ansible playbook to be performed
        """
        super().__init__(
            name=name,
            namespace=namespace,
            client=client,
            teardown=teardown,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )
        self.image = image
        self.playbook = playbook

    def to_dict(self):
        res = super().to_dict()
        if self.yaml_file:
            return res

        res.update(
            {
                "spec": {
                    "image": self.image,
                    "playbook": self.playbook,
                },
            }
        )
        return res



from ocp_resources.constants import TIMEOUT_4MINUTES
from ocp_resources.resource import Resource


class Namespace(Resource):
    """
    Namespace object, inherited from Resource.
    """

    api_version = Resource.ApiVersion.V1

    class Status(Resource.Status):
        ACTIVE = "Active"

    def __init__(
        self,
        name=None,
        client=None,
        teardown=True,
        label=None,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        **kwargs,
    ):
        super().__init__(
            name=name,
            client=client,
            teardown=teardown,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )
        self.label = label

    def to_dict(self):
        res = super().to_dict()
        if self.yaml_file:
            return res

        if self.label:
            res.setdefault("metadata", {}).setdefault("labels", {}).update(self.label)
        return res



# -*- coding: utf-8 -*-

from ocp_resources.constants import TIMEOUT_4MINUTES
from ocp_resources.logger import get_logger
from ocp_resources.persistent_volume_claim import PersistentVolumeClaim
from ocp_resources.resource import NamespacedResource, Resource
from ocp_resources.utils import TimeoutExpiredError, TimeoutSampler


LOGGER = get_logger(name=__name__)


class DataVolume(NamespacedResource):
    """
    DataVolume object.
    """

    api_group = NamespacedResource.ApiGroup.CDI_KUBEVIRT_IO

    class Status(NamespacedResource.Status):
        BLANK = "Blank"
        PVC_BOUND = "PVCBound"
        IMPORT_SCHEDULED = "ImportScheduled"
        ClONE_SCHEDULED = "CloneScheduled"
        UPLOAD_SCHEDULED = "UploadScheduled"
        IMPORT_IN_PROGRESS = "ImportInProgress"
        CLONE_IN_PROGRESS = "CloneInProgress"
        UPLOAD_IN_PROGRESS = "UploadInProgress"
        SNAPSHOT_FOR_SMART_CLONE_IN_PROGRESS = "SnapshotForSmartCloneInProgress"
        SMART_CLONE_PVC_IN_PROGRESS = "SmartClonePVCInProgress"
        UPLOAD_READY = "UploadReady"
        UNKNOWN = "Unknown"
        WAIT_FOR_FIRST_CONSUMER = "WaitForFirstConsumer"

    class AccessMode:
        """
        AccessMode object.
        """

        RWO = "ReadWriteOnce"
        ROX = "ReadOnlyMany"
        RWX = "ReadWriteMany"

    class ContentType:
        """
        ContentType object
        """

        KUBEVIRT = "kubevirt"
        ARCHIVE = "archive"

    class VolumeMode:
        """
        VolumeMode object
        """

        BLOCK = "Block"
        FILE = "Filesystem"

    class Condition:
        class Type:
            READY = "Ready"
            BOUND = "Bound"
            RUNNING = "Running"

        class Status(Resource.Condition.Status):
            UNKNOWN = "Unknown"

    def __init__(
        self,
        name=None,
        namespace=None,
        source=None,
        size=None,
        storage_class=None,
        url=None,
        content_type=ContentType.KUBEVIRT,
        access_modes=None,
        cert_configmap=None,
        secret=None,
        client=None,
        volume_mode=None,
        hostpath_node=None,
        source_pvc=None,
        source_namespace=None,
        multus_annotation=None,
        bind_immediate_annotation=None,
        preallocation=None,
        teardown=True,
        privileged_client=None,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        api_name="pvc",
        **kwargs,
    ):
        super().__init__(
            name=name,
            namespace=namespace,
            client=client,
            teardown=teardown,
            privileged_client=privileged_client,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )
        self.source = source
        self.url = url
        self.cert_configmap = cert_configmap
        self.secret = secret
        self.content_type = content_type
        self.size = size
        self.access_modes = access_modes
        self.storage_class = storage_class
        self.volume_mode = volume_mode
        self.hostpath_node = hostpath_node
        self.source_pvc = source_pvc
        self.source_namespace = source_namespace
        self.multus_annotation = multus_annotation
        self.bind_immediate_annotation = bind_immediate_annotation
        self.preallocation = preallocation
        self.api_name = api_name

    def to_dict(self):
        res = super().to_dict()
        if self.yaml_file:
            return res
        res.update(
            {
                "spec": {
                    "source": {self.source: {"url": self.url}},
                    self.api_name: {
                        "resources": {"requests": {"storage": self.size}},
                    },
                }
            }
        )
        if self.access_modes:
            res["spec"][self.api_name]["accessModes"] = [self.access_modes]
        if self.content_type:
            res["spec"]["contentType"] = self.content_type
        if self.storage_class:
            res["spec"][self.api_name]["storageClassName"] = self.storage_class
        if self.secret:
            res["spec"]["source"][self.source]["secretRef"] = self.secret.name
        if self.volume_mode:
            res["spec"][self.api_name]["volumeMode"] = self.volume_mode
        if self.source == "http" or "registry":
            res["spec"]["source"][self.source]["url"] = self.url
        if self.cert_configmap:
            res["spec"]["source"][self.source]["certConfigMap"] = self.cert_configmap
        if self.source == "upload" or self.source == "blank":
            res["spec"]["source"][self.source] = {}
        if self.hostpath_node:
            res["metadata"].setdefault("annotations", {}).update(
                {
                    f"{NamespacedResource.ApiGroup.KUBEVIRT_IO}/provisionOnNode": self.hostpath_node
                }
            )
        if self.multus_annotation:
            res["metadata"].setdefault("annotations", {}).update(
                {
                    f"{NamespacedResource.ApiGroup.K8S_V1_CNI_CNCF_IO}/networks": self.multus_annotation
                }
            )
        if self.bind_immediate_annotation:
            res["metadata"].setdefault("annotations", {}).update(
                {
                    f"{NamespacedResource.ApiGroup.CDI_KUBEVIRT_IO}/storage.bind.immediate.requested": "true"
                }
            )
        if self.source == "pvc":
            res["spec"]["source"]["pvc"] = {
                "name": self.source_pvc or "dv-source",
                "namespace": self.source_namespace or self.namespace,
            }
        if self.preallocation is not None:
            res["spec"]["preallocation"] = self.preallocation

        return res

    def wait_deleted(self, timeout=TIMEOUT_4MINUTES):
        """
        Wait until DataVolume and the PVC created by it are deleted

        Args:
        timeout (int):  Time to wait for the DataVolume and PVC to be deleted.

        Returns:
        bool: True if DataVolume and its PVC are gone, False if timeout reached.
        """
        super().wait_deleted(timeout=timeout)
        return self.pvc.wait_deleted(timeout=timeout)

    def wait(self, timeout=600, failure_timeout=120):
        self._check_none_pending_status(failure_timeout=failure_timeout)

        # If DV's status is not Pending, continue with the flow
        self.wait_for_status(status=self.Status.SUCCEEDED, timeout=timeout)
        self.pvc.wait_for_status(
            status=PersistentVolumeClaim.Status.BOUND, timeout=timeout
        )

    @property
    def pvc(self):
        return PersistentVolumeClaim(
            client=self.privileged_client or self.client,
            name=self.name,
            namespace=self.namespace,
        )

    @property
    def scratch_pvc(self):
        return PersistentVolumeClaim(
            name=f"{self.name}-scratch",
            namespace=self.namespace,
            client=self.client,
        )

    def _check_none_pending_status(self, failure_timeout=120):
        # Avoid waiting for "Succeeded" status if DV's in Pending/None status
        sample = None
        try:
            for sample in TimeoutSampler(
                wait_timeout=failure_timeout,
                sleep=10,
                func=lambda: self.instance.status.phase,
            ):
                # If DV status is Pending (or Status is not yet updated) continue to wait, else exit the wait loop
                if sample in [self.Status.PENDING, None]:
                    continue
                else:
                    break
        except TimeoutExpiredError:
            LOGGER.error(f"{self.name} status is {sample}")
            raise



from openshift.dynamic.exceptions import ResourceNotFoundError

from ocp_resources.constants import TIMEOUT_4MINUTES
from ocp_resources.logger import get_logger
from ocp_resources.persistent_volume_claim import PersistentVolumeClaim
from ocp_resources.resource import NamespacedResource


LOGGER = get_logger(name=__name__)


class DataSource(NamespacedResource):
    api_group = NamespacedResource.ApiGroup.CDI_KUBEVIRT_IO

    def __init__(
        self,
        name=None,
        namespace=None,
        client=None,
        source=None,
        teardown=True,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        **kwargs,
    ):
        super().__init__(
            name=name,
            namespace=namespace,
            client=client,
            teardown=teardown,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )
        self.source = source

    def to_dict(self):
        res = super().to_dict()
        if self.yaml_file:
            return res

        res.update(
            {
                "spec": {
                    "source": self.source,
                },
            }
        )

        return res

    @property
    def pvc(self):
        data_source_pvc = self.instance.spec.source.pvc
        pvc_name = data_source_pvc.name
        pvc_namespace = data_source_pvc.namespace
        try:
            return PersistentVolumeClaim(
                client=self.client,
                name=pvc_name,
                namespace=pvc_namespace,
            )
        except ResourceNotFoundError:
            LOGGER.warning(
                f"dataSource {self.name} is pointing to a non-existing PVC, name: {pvc_name}, "
                f"namespace: {pvc_namespace}"
            )



from ocp_resources.resource import NamespacedResource


class NodeHealthCheck(NamespacedResource):
    """
    NodeHealthCheck object.
    Reference : https://github.com/medik8s/node-healthcheck-operator
    """

    api_group = NamespacedResource.ApiGroup.REMEDIATION_MEDIK8S_IO



from ocp_resources.constants import NOT_FOUND_ERROR_EXCEPTION_DICT
from ocp_resources.logger import get_logger
from ocp_resources.resource import NamespacedResource
from ocp_resources.utils import TimeoutSampler


LOGGER = get_logger(name=__name__)


class Benchmark(NamespacedResource):
    """
    Benchmark resource
    Defined by https://github.com/cloud-bulldozer/benchmark-operator

    The benchmark-operator monitors a namespace for `Benchmark` resources.
    When a new `Benchmark` is created, the benchmark-operator creates and starts the pods or VMs necessary,
    and triggers the benchmark run.
    """

    api_group = NamespacedResource.ApiGroup.RIPSAW_CLOUDBULLDOZER_IO

    class Status:
        NONE = None  # None state is valid for newly created benchmark resources

    class Workload:
        class Kind:
            VM = "vm"
            POD = "pod"

    def _wait_for_instance_key(self, parent, key):
        """
        Wait for key to exist in parent attribute of instance

        Args:
            parent (str): An attribute of self.instance that should contain key
            key (str): A dictionary entry within parent

        Returns:
            str or None: Value of key if found, otherwise None
        """
        samples = TimeoutSampler(
            wait_timeout=30,
            sleep=1,
            func=lambda: getattr(self.instance, parent, None),
            exceptions_dict=NOT_FOUND_ERROR_EXCEPTION_DICT,
        )
        for sample in samples:
            if sample:
                return sample.get(key)

    @property
    def uuid(self):
        """
        Returns:
            str: UUID string from resource instance
        """
        return self._wait_for_instance_key(parent="status", key="uuid")

    @property
    def suuid(self):
        """
        Returns:
            str: (short)UUID string from resource instance
        """
        return self._wait_for_instance_key(parent="status", key="suuid")

    @property
    def workload_kind(self):
        """
        Retrieve the value of spec.workload.args.kind

        Not all Benchmarks have a 'kind' defined, this was added for vms. The default is 'pod'

        Returns:
            str: Value representing workload kind

        """
        return self.workload_arg(arg="kind", default="pod")

    def workload_arg(self, arg, default=None):
        """
        Retrieve the value of spec.workload.args[arg]

        To provide a similar usage as .get(), a default can be defined if needed.

        Args:
            arg (str): Argument to retrieve from spec.workload.args
            default (any): Default value to return if arg is not found in workload args

        Returns:
            any: Value of workload arg or 'default' if does not exist
        """
        workload = self._wait_for_instance_key(parent="spec", key="workload")
        if workload:
            return workload.get("args", {}).get(arg, default)
        return default



from ocp_resources.resource import NamespacedResource


class TektonTasks(NamespacedResource):
    """
    TektonTasks object, inherited from NamespacedResource.
    """

    api_group = NamespacedResource.ApiGroup.TEKTONTASKS_KUBEVIRT_IO



from ocp_resources.constants import TIMEOUT_4MINUTES
from ocp_resources.mtv import MTV
from ocp_resources.resource import NamespacedResource


class Provider(NamespacedResource, MTV):
    """
    Migration Toolkit For Virtualization (MTV) Provider object.
    """

    api_group = NamespacedResource.ApiGroup.FORKLIFT_KONVEYOR_IO

    def __init__(
        self,
        name=None,
        namespace=None,
        provider_type=None,
        url=None,
        secret_name=None,
        secret_namespace=None,
        vddk_init_image=None,
        client=None,
        teardown=True,
        yaml_file=None,
        delete_timeout=TIMEOUT_4MINUTES,
        **kwargs,
    ):
        super().__init__(
            name=name,
            namespace=namespace,
            client=client,
            teardown=teardown,
            yaml_file=yaml_file,
            delete_timeout=delete_timeout,
            **kwargs,
        )
        self.provider_type = provider_type
        self.url = url
        self.secret_name = secret_name
        self.secret_namespace = secret_namespace
        self.condition_message_ready = self.ConditionMessage.PROVIDER_READY
        self.vddk_init_image = vddk_init_image

    def to_dict(self):
        res = super().to_dict()
        if self.yaml_file:
            return res

        res.update(
            {
                "spec": {
                    "type": self.provider_type,
                    "url": self.url,
                    "secret": {
                        "name": self.secret_name,
                        "namespace": self.secret_namespace,
                    },
                    "settings": {"vddkInitImage": self.vddk_init_image},
                }
            }
        )

        return res



from ocp_resources.resource import NamespacedResource


class NetworkAttachmentDefinition(NamespacedResource):
    """
    NetworkAttachmentDefinition object.
    """

    api_group = NamespacedResource.ApiGroup.K8S_CNI_CNCF_IO
    resource_name = None

    def wait_for_status(
        self, status, timeout=None, label_selector=None, resource_version=None
    ):
        raise NotImplementedError(f"{self.kind} does not have status")

    def to_dict(self):
        res = super().to_dict()
        if self.yaml_file:
            return res

        if self.resource_name is not None:
            res["metadata"]["annotations"] = {
                f"{NamespacedResource.ApiGroup.K8S_V1_CNI_CNCF_IO}/resourceName": self.resource_name
            }
        res["spec"] = {}
        return res



# -*- coding: utf-8 -*-
from ocp_resources.constants import PROTOCOL_ERROR_EXCEPTION_DICT, TIMEOUT_4MINUTES
from ocp_resources.logger import get_logger
from ocp_resources.resource import NamespacedResource
from ocp_resources.utils import TimeoutSampler


LOGGER = get_logger(name=__name__)


class Deployment(NamespacedResource):
    """
    OpenShift Deployment object.
    """

    api_group = NamespacedResource.ApiGroup.APPS

    def scale_replicas(self, replica_count=int):
        """
        Update replicas in deployment.

        Args:
            replica_count (int): Number of replicas.

        Returns:
            Deployment is updated successfully
        """
        body = super().to_dict()
        body.update({"spec": {"replicas": replica_count}})

        LOGGER.info(f"Set deployment replicas: {replica_count}")
        return self.update(resource_dict=body)

    def wait_for_replicas(self, deployed=True, timeout=TIMEOUT_4MINUTES):
        """
        Wait until all replicas are updated.

        Args:
            deployed (bool): True for replicas deployed, False for no replicas.
            timeout (int): Time to wait for the deployment.

        Raises:
            TimeoutExpiredError: If not availableReplicas is equal to replicas.
        """
        LOGGER.info(f"Wait for {self.kind} {self.name} to be deployed: {deployed}")
        samples = TimeoutSampler(
            wait_timeout=timeout,
            sleep=1,
            exceptions_dict=PROTOCOL_ERROR_EXCEPTION_DICT,
            func=lambda: self.instance,
        )
        for sample in samples:
            if sample:
                status = sample.status

                spec_replicas = sample.spec.replicas
                total_replicas = status.replicas or 0
                updated_replicas = status.updatedReplicas or 0
                available_replicas = status.availableReplicas or 0
                ready_replicas = status.readyReplicas or 0

                if (
                    (deployed and spec_replicas)
                    and spec_replicas
                    == updated_replicas
                    == available_replicas
                    == ready_replicas
                ) or not (deployed or spec_replicas or total_replicas):
                    return



from ocp_resources.resource import NamespacedResource


class InstallPlan(NamespacedResource):
    api_group = NamespacedResource.ApiGroup.OPERATORS_COREOS_COM

    class Status(NamespacedResource.Status):
        COMPLETE = "Complete"
